/*
 * Copyright (c) 2020 RTSoft. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @file ftactions.c
 * @brief Different operations with devices
 */

#include "fttransport.h"
#include "ftlibtypes.h"
#include "ftconfig.h"

#include <stdio.h>
#include <math.h>

/**
 @brief Get TXT2 for device
 @param [in] device device connected to TXT2 controller
 @return TXT2 device
 */
extern void *ftop_get_txt(void *device);

/**
 @brief Debug procedure to output I1-I8 config
 @param [in] iconfig config info
 */
void ftOutputInputsConfig(struct ft_inputs_config *iconfig);

/**
 @brief Debug procedure to output M1-M4 config
 @param [in] mconfig config info
 */
void ftOutputOutputsConfig(struct ft_motors_config *mconfig);

extern int ftdebug; /*!< common debug mask */

/**
 * @brief Start motor operation
 * @param txt2 TXT2 structure
 * @param action motor action
 * @param motor motor comection
 * @param speed new speed
 * @param counter connected counter
 * @param distance new distance
 * @param cmdid new cmdid
 */
static void ftmop_start_motor(ftobject_txt2 *txt2, int action, int motor,
		int speed, int counter, int distance, int cmdid)
{
	struct ftop_start_motor command = { { sizeof(command), ftop_op_start_motor,
			txt2->canid, 0 }, .action = action, .motor1 = motor,
			.speed1 = speed, .counter1 = counter, .distance = distance,
			.cmdid1 = cmdid };
	ftop_socket_send(txt2->transport, &command, sizeof(command));
}

/**
 * @brief Start motor operation
 * @param txt2 TXT2 structure
 * @param action motor action
 * @param motor1 first motor comection
 * @param speed1 first motor speed
 * @param counter1 first motor counter
 * @param motor2 second motor comection
 * @param speed2 second motor speed
 * @param counter2 second motor counter
 * @param distance new distance
 * @param cmdid1 first motor cmdid
 * @param cmdid2 second motor cmdid
 */
static void ftmop_start_motors(ftobject_txt2 *txt2, int action, int motor1,
		int speed1, int counter1, int motor2, int speed2, int counter2,
		int distance, int cmdid1, int cmdid2)
{
	struct ftop_start_motor command = {
			{ sizeof(command), ftop_op_start_motor, txt2->canid, 0 },
			.action = action, .motor1 = motor1,
			.speed1 = speed1, .counter1 = counter1, .distance = distance,
			.motor2 = motor2, .speed2 = speed2, .counter2 = counter2,
			.cmdid1 = cmdid1, .cmdid2 = cmdid2 };
	ftop_socket_send(txt2->transport, &command, sizeof(command));
}

/**
 @brief Get resistance value
 @param [in] device resistor device
 @return resistance
 */
int ftop_get_resistance(void *device)
{
	/* I1-I8 */
	ftobject_connection *obj = device;
	ftobject_txt2 *txt2 = ftop_get_txt(device);
	int conn = obj->conn - ftdev_i1;
	uint32_t voltage = txt2->cvalues->inputs[conn];

	if (ftdebug & FT_DEBUG_A7_CONVERT) {
	    fprintf(stderr, "voltage %d\n", voltage);
	}

	/* Calibration values */
	long long vh;
	long vl;
	if (txt2->version) {
		vh = voltage * 1830545LL;
		vl = 2861767 - 615 * voltage;
	} else {
		vh = voltage * 1820240LL;
		vl = 2856076 - 600 * voltage;
	}

	if (ftdebug & FT_DEBUG_A7_CONVERT) {
	    fprintf(stderr, "vh %lld vl %ld\n", vh, vl);
	}

	if (vl <= 0)
		return 0x7fffffff;

	int result = vh / vl - 100;

	if (ftdebug & FT_DEBUG_A7_CONVERT) {
	    fprintf(stderr, "result %d\n", result);
	}

	if (result < 0)
		result = 0x7fffffff;

	return result;
}

/**
 @brief Get voltage value
 @param [in] device Voltmeter device
 @return voltage mv
 */
int ftop_get_voltage(void *device)
{
    /* I1-I8 */
    ftobject_connection *obj = device;
    ftobject_txt2 *txt2 = ftop_get_txt(device);

    return txt2->cvalues->inputs[obj->conn - ftdev_i1];
}

/**
 @brief Get temperature value
 @param [in] device NCTTermistor device
 @return temperature
 */
int ftop_get_temperature(void *device)
{
	/* I1-I8 */
	/*
	 *      r = self.resistance
	 if r != 0:
	 x = log(r)
	 y = x * x * 1.39323522
	 z = x * -43.9417405
	 T = y + z + 271.870481
	 else:
	 T = 10000
	 return T
	 */
	int r = ftop_get_resistance(device);
	if (!r)
		return 10000000;

	double x = log(r);
	double y = x * x * 1.39323522;
	double z = x * -43.9417405;
	double T = y + z + 271.870481;

	return T * 1000;
}

/**
 @brief Get state value
 @param [in] device analog, digital or counter input device
 @return state (voltage for analog, current state for digital, counted value for counter)
 */
int ftop_get_state(void *device)
{
	/* I1-I8 + C1-C4*/
	ftobject_connection *obj = device;
	ftobject_txt2 *txt2 = ftop_get_txt(device);
	int conn = obj->conn;

	if (conn >= ftdev_i1 && conn <= ftdev_i8) {
		uint16_t val = txt2->cvalues->inputs[obj->conn - ftdev_i1];
		return (obj->magic == ft_switch_magic) ? !val : val;
	} else if (conn >= ftdev_c1 && conn <= ftdev_c4) {
		return !txt2->cvalues->cstate[conn - ftdev_c1];
	} else {
		return 0;
	}
}

/**
 @brief Get distance value
 @param [in] device ultrasonic input device
 @return distance
 */
int ftop_get_distance(void *device)
{
    /* I1-I8 */
    ftobject_connection *obj = device;
    ftobject_txt2 *txt2 = ftop_get_txt(device);
    int conn = obj->conn;

	if (conn >= ftdev_i1 && conn <= ftdev_i8) {
		return txt2->cvalues->inputs[obj->conn - ftdev_i1];
	} else if (conn >= ftdev_c1 && conn <= ftdev_c4) {
		return txt2->cvalues->counters[conn - ftdev_c1];
	} else {
		return 0;
	}
}

/**
 @brief Get values
 @param [in] device TXT2 device
 @param [out] values array of values (voltage, digital value or distance for ultrasonic)
 @return status
 */
int ftop_get_values(void *device, int values[FT_MAX_INPUTS + FT_MAX_COUNTERS])
{
	ftobject_txt2 *txt2 = device;

	for (int i = 0; i < FT_MAX_INPUTS; i++) {
		values[i] = txt2->cvalues->inputs[i];
	}

	for (int i = 0; i < FT_MAX_COUNTERS; i++) {
		values[i + FT_MAX_INPUTS] = txt2->cvalues->counters[i];
	}

	return 0;
}

/**
 @brief Get faults info for TXT2
 @param [in] device TXT2 device
 @param [in] finfo output structure
 @return op status
 */
int ftop_get_finfo(void *device, struct ft_faults *finfo)
{
	ftobject_txt2 *txt2 = device;

	finfo->ocp = txt2->cvalues->ft_finfo.ocp;
	finfo->old = txt2->cvalues->ft_finfo.old;
	finfo->flags = txt2->cvalues->ft_finfo.flags;
	finfo->faults = txt2->cvalues->ft_finfo.faults;
	return 0;
}

/**
 @brief Set lamp brightness
 @param [in] device lamp device
 @param [in] value brightness (0-512)
 */
void ftop_set_brightness(void *device, int value)
{
	ftobject_txt2 *txt2 = ftop_get_txt(device);
	ftobject_motor *cobj = device;
	ft_pwm_type pwm = value;

	if (pwm > 512)
		pwm = 512;

	if (cobj->conn < ftdev_o1 || cobj->conn > ftdev_o8)
		return;

	cobj->speed = value;

	struct ftop_set_pwm command = {
		{sizeof(command), ftop_op_set_pwm, txt2->canid, 0},
		cobj->conn, pwm};

	ftop_socket_send(txt2->transport, &command, sizeof(command));

	if (ftdebug & FT_DEBUG_A7_OP) {
	    fprintf(stderr, "set pwm %d 0x%x\n", cobj->conn, pwm);
	}
}

/**
 @brief Set output PWM
 @param [in] device  device
 @param [in] value PWM (0-512)
 */
void ftop_set_pwm(void *device, int value)
{
	ftobject_txt2 *txt2 = ftop_get_txt(device);
	ftobject_motor *cobj = device;
	ft_pwm_type pwm = value;

	if (pwm > 512)
		pwm = 512;

	if (cobj->conn < ftdev_o1 || cobj->conn > ftdev_s3)
		return;

	if (cobj->conn >= ftdev_c1 && cobj->conn <= ftdev_c4)
		return;

	cobj->speed = value;

	struct ftop_set_pwm command = {
			{sizeof(command), ftop_op_set_pwm, txt2->canid, 0},
			cobj->conn, pwm};

	ftop_socket_send(txt2->transport, &command, sizeof(command));

	if (ftdebug & FT_DEBUG_A7_OP) {
	    fprintf(stderr, "set pwm %d 0x%x\n", cobj->conn, pwm);
	}
}

/**
 @brief Get saved output PWM
 @param [in] device  device
 @return PWM (0-512) value
 */
unsigned int  ftop_get_pwm(void *device)
{
	ftobject_motor *cobj = device;

	if (cobj->conn < ftdev_o1 || cobj->conn > ftdev_s3)
		return 0;

	if (cobj->conn >= ftdev_c1 && cobj->conn <= ftdev_c4)
		return 0;

	return cobj->speed;
}

/**
 @brief Set motor step counter
 @param [in] device motor device
 @param [in] counter step counter device
 */
void ftop_motor_set_step_counter(void *device, void *counter)
{
    ftobject_connection *cobj = counter;
    ftobject_motor *mobj = device;
    mobj->counter = cobj->conn;
}

/**
 @brief Set motor speed for the next operation
 @param [in] device motor device
 @param [in] speed new speed in range [-512, 512]
 */
void ftop_motor_set_speed(void *device, int speed)
{
    ftobject_motor *mobj = device;
    mobj->speed = speed;
}

/**
 @brief Immediatelly change speed and start motor
 @param [in] device motor device
 @param [in] speed new speed in range [-512, 512]
 */
void ftop_motor_start_speed(void *device, int speed)
{
    ftobject_txt2 *txt2 = ftop_get_txt(device);
    ftobject_motor *mobj = device;

    mobj->speed = speed;
    ftmop_start_motor(txt2, ftop_m_start, mobj->conn, speed, mobj->counter, 0, ++mobj->cmdid);
}

/**
 @brief Immediatelly change speed and start motor
 @param [in] device motor device
 @param [in] distance distance
 @param [in] device2 synchronized motor device (may be NULL)
 */
void ftop_motor_start_distance(void *device, int distance, void *device2)
{
    ftobject_txt2 *txt2 = ftop_get_txt(device);
    ftobject_motor *mobj = device;

    if (device2 != NULL) {
        ftobject_motor *mobj2 = device2;
        ftmop_start_motors(txt2, ftop_m_distance_sync, mobj->conn, mobj->speed, mobj->counter,
            mobj2->conn, mobj2->speed, mobj2->counter, distance, ++mobj->cmdid, ++mobj2->cmdid);
    } else {
        ftmop_start_motor(txt2, ftop_m_distance, mobj->conn, mobj->speed, mobj->counter, distance, ++mobj->cmdid);
    }
}

/**
 @brief Immediatelly change speed and start motor
 @param [in] device motor device
 @param [in] distance distance, 0 - unlimited
 @param [in] mask 4 bit mask for motors
 */
void ftop_motor_start_sync_distance(void *device, int distance, int mask)
{
	ftobject_txt2 *txt2 = ftop_get_txt(device);
	ftobject_motor *mobj = device;
	if (mobj->conn < ftdev_m1 || mobj->conn > ftdev_m4)
		return;

	mask |= 1 << ((mobj->conn - ftdev_m1) / 2);
	struct ftop_start_sync_motor command = {
			{ sizeof(command), ftop_op_start_sync_motor, txt2->canid, 0 },
			.action = ftop_m_distance_sync, .mask = mask, .distance = distance
		};

	/* Collect motor specific information */
	for (int i = 0; i < FT_MAX_MOTORS; i++) {
		command.speed[i] = txt2->outputs[i * 2].motor.speed;
		if (mask & (1 << i))
			txt2->outputs[i * 2].motor.cmdid++;
		command.cmdid[i] = txt2->outputs[i * 2].motor.cmdid;
	}

	ftop_socket_send(txt2->transport, &command, sizeof(command));
}

/**
 @brief Immediatelly change speed and start motor
 @param [in] device motor device
 @param [in] mask 4 bit mask for motors
 */
void ftop_motor_stop_sync(void *device, int mask)
{
	ftobject_txt2 *txt2 = ftop_get_txt(device);
	ftobject_motor *mobj = device;
	if (mobj->conn < ftdev_m1 || mobj->conn > ftdev_m4)
		return;

	mask |= 1 << ((mobj->conn - ftdev_m1) / 2);
	struct ftop_start_sync_motor command = {
			{ sizeof(command), ftop_op_start_sync_motor, txt2->canid, 0 },
			.action = ftop_m_stop, .mask = mask
		};

	/* Collect motor specific information */
	for (int i = 0; i < FT_MAX_MOTORS; i++) {

		if (mask & (1 << i)) {
			txt2->outputs[i * 2].motor.cmdid++;
			txt2->outputs[i * 2].motor.speed = 0;
		}
		command.speed[i] = txt2->outputs[i * 2].motor.speed;
		command.cmdid[i] = txt2->outputs[i * 2].motor.cmdid;
	}

	ftop_socket_send(txt2->transport, &command, sizeof(command));
}

/**
 @brief Immediatelly stop motor
 @param [in] device motor device
 */
void ftop_motor_stop(void *device)
{
	ftobject_txt2 *txt2 = ftop_get_txt(device);
	ftobject_motor *mobj = device;

	mobj->speed = 0;
	ftmop_start_motor(txt2, ftop_m_stop, mobj->conn, mobj->speed, mobj->counter,
			0, ++mobj->cmdid);
}

/**
 @brief Reset device, for example, counter
 @param [in] device motor/counter device
 */
void ftop_zero(void *device)
{
	ftobject_txt2 *txt2 = ftop_get_txt(device);
	ftobject_connection *cobj = device;

	struct ftop_set_pwm command = {
				{sizeof(command), ftop_op_zero, txt2->canid, 0},
			cobj->conn, 0};
	ftop_socket_send(txt2->transport, &command, sizeof(command));
}

/**
 @brief Get distance value for motor
 @param [in] device motor device
 @return distance
 */
int ftop_motor_get_distance(void *device)
{
    ftobject_txt2 *txt2 = ftop_get_txt(device);
    ftobject_motor *mobj = device;

    return txt2->cvalues->counters[(mobj->conn - ftdev_m1) / 2];
}

/**
 @brief Get running state for the motor
 @param [in] device the motor device
 @return motor is running
 */
_Bool ftop_motor_isrunning(void *device)
{
    ftobject_txt2 *txt2 = ftop_get_txt(device);
    ftobject_motor *mobj = device;
    int m =  (mobj->conn - ftdev_m1) / 2;

    return txt2->current_values.cmdvalues.cmdids[m] != mobj->cmdid;
}

/**
 @brief Set CAN speed
 @param [in] device CAN device
 @param [in] speed new speed
 */
void ftop_can_set_bitrate(void *device, int speed)
{
	ftobject_txt2 *txt2 = device;
	struct ftop_set_mode command = {
			{sizeof(command), ftop_op_set_bitrate, txt2->canid, 0},
			speed};
	ftop_socket_send(txt2->transport, &command, sizeof(command));
}

/**
 @brief Wait txt2 controller event
 @param [in] device TXT2 controller
 @param [in] emask event mask (enum ft_bevent)
 @return event mask
 */
int ftop_wait_event(void *device, uint32_t emask)
{
	ftobject_txt2 *txt2 = device;
	int result = 0;

	if (!(emask & ((1 << ft_event_max) - 1)))
		return 0;

	pthread_mutex_lock(&txt2->emutex);

	struct fttxt_events fte = txt2->events;
	do {

		pthread_cond_wait(&txt2->econd, &txt2->emutex);
		for (int i = 0; i < ft_event_max; i++) {
			if ((emask & (1 << i)) && fte.counts[i] != txt2->events.counts[i]) {
				result |= 1 << i;
			}
		}
	} while(!result);

	pthread_mutex_unlock(&txt2->emutex);

	return result;
}

/**
 @brief Update config
 @param [in] device TXT2 controller
 */
void ftop_update_config(void *device)
{
    ftobject_txt2 *txt2 = device;
    struct ftop_setup_config command = {
    		{sizeof(command), ftop_op_send_config, txt2->canid, 0},
			txt2->connusage, txt2->new_config};

    ftop_socket_send(txt2->transport, &command, sizeof(command));

	if (ftdebug & FT_DEBUG_A7_OP) {
		fprintf(stderr, "update_config 0x%x\n", txt2->connusage);
		ftOutputInputsConfig(&txt2->new_config.inputs);
		ftOutputOutputsConfig(&txt2->new_config.motors);
	}
}

/**
 @brief Power down all TXT2 devices
 @param [in] device TXT2 controller
 */
void ftop_power_down(void *device)
{
	ftobject_txt2 *txt2 = device;
	struct ftop_command_packet command = { sizeof(command), ftop_op_power_down,
			txt2->canid, 0 };

	ftop_socket_send(txt2->transport, &command, sizeof(command));
}

/**
 @brief Reset all TXT2 defices
 @param [in] device TXT2 controller
 */
void ftop_reset(void *device)
{
	ftobject_txt2 *txt2 = device;
	struct ftop_command_packet command = {sizeof(command), ftop_op_reset, txt2->canid, 0};

	ftop_socket_send(txt2->transport, &command, sizeof(command));
}

/**
 @brief Set coast mode for the motors
 @param [in] device TXT2 device
 @param [in] mask for motors
 */
void ftop_motor_coast(void *device, int mask)
{
	ftobject_txt2 *txt2 = device;
	struct ftop_set_mode command = { { sizeof(command), ftop_op_motor_coast,
			txt2->canid, 0 }, mask };

	ftop_socket_send(txt2->transport, &command, sizeof(command));
}
