/*
 * Copyright (c) 2020 RTSoft. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @file rpmsgTransport.c
 * @brief Linux side rpmsg communication procedures
 *
 * Client transport
 */

#include "../../inc/ftconfig.h"

#include "fttransport.h"

#include <malloc.h>
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <termios.h>
#include <fcntl.h>
#include <stdint.h>
#include <sys/mman.h>
#include <sys/stat.h>

extern int ftdebug;

/**
 * Transport structure
 */
typedef ftop_transport stransport;

/**
 * @brief Send packet to M4
 * @param transport transport structure
 * @param packet packet with command
 * @param size command size
 * @return packet length
 */
int ftop_socket_send(void *transport, void *packet, int size)
{
    stransport *s = transport;
    usleep(10); // TODO: fix it
    ssize_t len = write(s->fid, packet, size);
    if (ftdebug & FT_DEBUG_A7_TRANSPORT) {
		fprintf(stderr, "send bytes %d\n", len);
    }
    return len;
}

/**
 * @brief Receive packet from M4
 * @param transport transport structure
 * @param packet buffer for packet
 * @param len buffer length
 * @return packet length
 */
int ftop_socket_receive(void *transport, void *packet, int len)
{
    stransport *s = transport;
    int tlen = read(s->fid, packet, len);

    if (ftdebug & FT_DEBUG_A7_TRANSPORT) {
		fprintf(stderr, "receive bytes %d\n", tlen);
    }

    return tlen;
}

/**
 * @brief Free communication channel
 * @return success 0
 */
int ftop_socket_free(void *transport)
{
    stransport *s = transport;

    close(s->fid);
    close(s->lfid);
    free(s);

    return 0;
}

/**
 * @brief Get lock file id
 * @return lock file id
 */
int ftop_open_lock()
{
	mode_t mask = umask(0);
	int lid = open("/run/lock/LCK..ftrpmsg", O_RDWR | O_CREAT | O_CLOEXEC, 0666);
	umask(mask);
	if (lid < 0) {
	    perror("Can't open lock file");
	}
	return lid;
}

/**
 * @brief Try to lock client
 * @param lid lock file
 * @param client client id
 * @return success or fail
 */
int ftop_lock_client(int lid, int client)
{
	struct flock lock = {.l_type = F_WRLCK, .l_whence = SEEK_SET, .l_len = 1, .l_start = client};

	return fcntl(lid, F_SETLK, &lock);
}

/**
 * @brief Try to unlock client
 * @param lid lock file
 * @param client client id
 * @return success or fail
 */
int ftop_unlock_client(int lid, int client)
{
	struct flock lock = {.l_type = F_UNLCK, .l_whence = SEEK_SET, .l_len = 1, .l_start = client};

	return fcntl(lid, F_SETLK, &lock);
}

/**
 * @brief Try to lock client
 * @param client client id
 * @return file id
 */
int ftop_open_client_file(int client)
{
	char rpmsgname[12] = "/dev/rpmsg0";

	rpmsgname[10] = '0' + client;
	int id = open (rpmsgname, O_RDWR | O_CLOEXEC);

	return id;
}

/**
 * @brief Get transport
 * @param host unused
 * @parm port unused
 * @return transport structure
 */
ftop_transport* ftop_socket(const char *host, int port)
{
	(void)host; (void)port;

	int lid = ftop_open_lock();
	if (lid < 0)
		return NULL;

	for (int i = 0; i < 8; i++) {
		if (ftop_lock_client(lid, i) != -1) {
			int sync_fd = ftop_open_client_file(i);
			if (sync_fd != -1) {
				stransport *s = calloc(1, sizeof(s[0]));
			    s->fid = sync_fd;
			    s->lfid = lid;
			    s->cid = i;
			    return s;
			}
			ftop_unlock_client(lid, i);
		}
	}

	return NULL;
}

