/*
 * Copyright (c) 2020 RTSoft. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @file tmix.cpp
 * @brief FT API C++ example
 *
 * C++ mixed API test
 */

#include <iostream>
#include <string>
#include <stdio.h>
#include <unistd.h>

extern "C" {
#include "ftlock.h"
}
#include "ft.hpp"

using namespace ft;

static void Usage(int code)
{
	std::cerr << "Usage: tmix [test_name] [pin 1-8]" << std::endl
			  << "lamp         - lamp test, default pin O7" << std::endl
			  << "servo        - servo test, default pin S1" << std::endl
			  << "motor        - motor test, default pins M3" << std::endl
			  << "encoder      - encoder motor test, default pins M1" << std::endl
			  << "resistor     - resistor test, default pin I4" << std::endl
			  << "ncttermistor - ncttermistor test, default pin I5" << std::endl
			  << "switch       - switch test, default pin I8" << std::endl
			  << "digital      - digital test, default pin I6" << std::endl
			  << "voltmeter    - voltmeter test, default pin I2" << std::endl
			  << "ultrasonic   - ultrasonic test, default pin I1" << std::endl
			  << "counter      - counter test, default pin C1" << std::endl;

	exit(code);
}

static int getInput(char *argv, int def)
{
	if (argv == 0)
		return def;

	int result = def;
	sscanf(argv, "%i", &result);
	return result;
}

static void LampTest(TXT& txt2, int output)
{
	Lamp l { txt2, output };
	txt2.update_config();
	for (int i = 0; i <= 512; i += 20) {
		l.setBrightness(i);
		sleep(1);
	}
	l.setBrightness(0);
}

static void ServoTest(TXT& txt2, int output)
{
	Servo s { txt2, output };
	txt2.update_config();
	for (int i = 0; i <= 512; i += 20) {
		s.setPwm(i);
		sleep(1);
	}
}

static void MotorTest(TXT& txt2, int output)
{
	Motor m { txt2, output };
	txt2.update_config();
	m.stop();
	m.startSpeed(256);
	sleep(1);
	m.startSpeed(0);
	sleep(1);
	m.startSpeed(-256);
	sleep(1);
	m.startSpeed(0);
}

static void BrakeTest(Encoder& e, int speed, int distance)
{
	e.setSpeed(speed);
	e.startDistance(distance);
	while(e.isRunning());
	std::cerr << "Speed " << speed << " distance " << distance
			<< " result position " << e.getCurrentDistance() << std::endl;
}

static void EncoderTest(TXT& txt2, int output)
{
	static struct {
		int speed;
		int distance;
	} subtests[] = {
			{150, 10},
			{150, 20},
			{150, 25},
			{150, 26},
			{150, 266},
			{400, 266},
			{400, 26},
			{512, 266}

	};

	Encoder e { txt2, output };
	txt2.update_config();

	for(const auto& t : subtests) {
		BrakeTest(e, t.speed, t.distance);
	}
}

static void ResistorTest(TXT& txt2, int input)
{
	Resistor r {txt2, input};
	txt2.update_config();

	std::cerr << "Resistance : " << r.getResistance() << std::endl;
	std::cerr << "State      : " << r.getState() << std::endl;
}

static void NCTTermistorTest(TXT& txt2, int input)
{
	NCTTermistor t { txt2, input };
	txt2.update_config();

	while (true) {
		std::cerr << "Temperature : " << t.getTemperature() << std::endl;
		std::cerr << "Resistance  : " << t.getResistance() << std::endl;
		std::cerr << "State       : " << t.getState() << std::endl;
	}
}

static void SwitchTest(TXT& txt2, int input)
{
	Switch s { txt2, input };
	txt2.update_config();

	while (true) {
		std::cerr << "State      : " << s.getState() << std::endl;
	}
}

static void DigitalTest(TXT& txt2, int input)
{
	Digital d { txt2, input };
	txt2.update_config();

	while (true) {
		std::cerr << "State      : " << d.getState() << std::endl;
	}
}

static void VoltmeterTest(TXT& txt2, int input)
{
	Voltmeter v { txt2, input };
	txt2.update_config();

	while (true) {
		std::cerr << "State      : " << v.getState() << std::endl;
	}
}

static void UltrasonicTest(TXT& txt2, int input)
{
	Ultrasonic u { txt2, input };
	txt2.update_config();

	while (true) {
		std::cerr << "Distance : " << u.getDistance() << std::endl;
	}
}

static void CounterTest(TXT& txt2, int input)
{
	Counter c { txt2, input };
	int pstate = -1, distance = -1;

	txt2.update_config();

	while (true) {
		int nstate = c.getState();

		if (nstate != pstate) {
			std::cerr << "State      : " << nstate << std::endl;
			pstate = nstate;
		}

		int ndistance = c.getDistance();

		if (ndistance != distance) {
			std::cerr << "Distance   : " << ndistance << std::endl;
			distance = ndistance;
		}

		if (ndistance > 10) {
			std::cerr << "Reset" << std::endl;
			c.reset();
		}
	}
}

static void execute(std::string& command, char *arg)
{
	TXT txt2;

	ftlock_notify('R');

	if (command == "lamp") {
		LampTest(txt2, getInput(arg, 7));
	} else if (command == "servo") {
		ServoTest(txt2, getInput(arg, 1));
	} else if (command == "motor") {
		MotorTest(txt2, getInput(arg, 3));
	} else if (command == "encoder") {
		EncoderTest(txt2, getInput(arg, 1));
	} else if (command == "resistor") {
		ResistorTest(txt2, getInput(arg, 4));
	} else if (command == "ncttermistor") {
		NCTTermistorTest(txt2, getInput(arg, 5));
	} else if (command == "switch") {
		SwitchTest(txt2, getInput(arg, 8));
	} else if (command == "digital") {
		DigitalTest(txt2, getInput(arg, 6));
	} else if (command == "voltmeter") {
		VoltmeterTest(txt2, getInput(arg, 2));
	} else if (command == "ultrasonic") {
		UltrasonicTest(txt2, getInput(arg, 1));
	} else if (command == "counter") {
		CounterTest(txt2, getInput(arg, 1));
	} else {
		Usage(1);
	}
}

/**
 * @brief Standard main procedure
 * @param [in] argc number of arguments
 * @param [in] argv arguments
 * @return Error code
 */
int main(int argc, char **argv)
{
	if (argc <= 1)
		Usage(0);

        if (ftlock_open() < 0)
		return 1;

	if (ftlock_try_lock(argv[0]) < 0)
		return 1;

	std::string command{argv[1]};
        execute(command, argv[2]);

	ftlock_unlock(); // can be skipped in this case
	ftlock_close();  // can be skipped in this case

	return 0;
}
