/*
 * Copyright (c) 2020 RTSoft. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/*!
  @file  ftimpltypes.h
  @brief Implementation specific types
 */

#ifndef _FT_IMPL_TYPES
#define _FT_IMPL_TYPES

#define FT_INTERNAL

#include "fttypes.h"

/** Input config */
struct ft_inputs_config {
	uint8_t inputs[FT_MAX_INPUTS][2]; /*!< I1-I8 configuration */
};

/** O1-O8/M1-M4 pair modes */
enum ft_outputs_usage {
    ft_output_mode,
    ft_motor_mode,
};

/** Set of motor modes */
struct ft_motors_config {
    uint8_t usage[FT_MAX_MOTORS]; /*!< O1-O8 usage */
};

/** TXT2 input/output configuration */
struct ft_config {
    struct ft_inputs_config inputs; /*!< I1-I8 configuration */
    struct ft_motors_config motors; /*!< M1-M4 configuration */
};

/** Motor actions */
enum ftop_m_actions {
    ftop_m_unknown,       /*!< nop */
    ftop_m_start,         /*!< start one with speed */
    ftop_m_distance,      /*!< start one with distance */
    ftop_m_start_sync,    /*!< start synchronous with speed */
    ftop_m_distance_sync, /*!< start synchronous with distance */
    ftop_m_stop,          /*!< stop */
};

/** Commands */
enum ftop_commands {
    ftop_op_ping,        /*!< ping op value */
    ftop_op_set_mode,    /*!< set TXT2 mode */
    ftop_op_get_info,    /*!< request TXT2 info */
    ftop_op_get_config,  /*!< request config */
    ftop_op_send_config, /*!< setup config */
    ftop_op_get_values,  /*!< get current values */
    ftop_op_set_pwm,     /*!< set pwm for O1-O8, S1-S3 */
    ftop_op_start_motor, /*!< start motor */
	ftop_op_debug,       /*!< debug output */
	ftop_op_disconnect,  /*!< disconnect */
	ftop_op_terminator,  /*!< set FDCAN terminator */
	ftop_op_set_uid,     /*!< set uid for TXT */
	ftop_op_set_name,    /*!< set name for master */
	ftop_op_set_bitrate, /*!< set FDCAN bitrate */
	ftop_op_power_down,  /*!< stio threads, powerdown devices and stop */
	ftop_op_reset,       /*!< reset devices to predefined state */
	ftop_op_start_sync_motor, /*!< start sync motor */
	ftop_op_zero,        /*!< Set zero state (counters and so on) */
	ftop_op_nresponse,   /*!< New values response */
	ftop_op_nping,		 /*!< New ping op value */
	ftop_op_motor_coast, /*!< Set motor coast mode */
};

/** Command packet header */
struct ftop_command_packet {
    uint8_t len;     /*!< packet len */
    uint8_t cmd;     /*!< command code */
    uint8_t dst;     /*!< destination TXT2 code */
    uint8_t cmdid;   /*!< command id */
};

/** Config setup packet */
struct ftop_set_mode {
    struct ftop_command_packet packet; /*!< common command header */
    int mode;                          /*!< new mode */
};

/** Name setup packet */
struct ftop_set_name {
    struct ftop_command_packet packet; /*!< common command header */
    char name[FT_NAME_SIZE];           /*!< new name */
};

/** Unique STM32MP157 id */
typedef uint32_t ftuid;

/** TXT2 info response packet */
struct ftop_response_txt_info {
    struct ftop_command_packet packet; /*!< common response header */
    int version;			 /*!< board revision    */
    int firmware;			 /*!< firmware verision */
    ftuid uid;				 /*!< board uid */
    int mode;				 /*!< board mode */
    int slaves;				 /*!< number of slaves */
    uint32_t uaddr;			 /*!< initial address for memory mapping */
    uint32_t taddr;			 /*!< taddress for memory mapping */
    uint32_t vaddr;			 /*!< values address for memory mapping */
    char name[FT_NAME_SIZE]; /*!< txt2 name */
};

/** Config response packet */
struct ftop_response_config {
    struct ftop_command_packet packet; /*!< common response header */
    int configid;                      /*!< config id */
    struct ft_config config;           /*!< config itself */
};

/** Config setup packet */
struct ftop_setup_config {
    struct ftop_command_packet packet; /*!< common command header */
    uint32_t usage;                    /*!< resources usage */
    struct ft_config config;           /*!< config itself */
};

/** PWM setup packet */
struct ftop_set_pwm {
    struct ftop_command_packet packet; /*!< common command header */
    uint8_t devid;                     /*!< device id */
    ft_pwm_type pwm;                   /*!< new PWM value 0-512 */
};

/** Start motor packet */
struct ftop_start_motor {
    struct ftop_command_packet packet; /*!< common command header */
    uint16_t distance;                 /*!< distance */
    uint16_t  cmdid1;                  /*!< first motor cmdid */
    int16_t  speed1;                   /*!< first motor signed speed */
    uint8_t action;                    /*!< command action */
    uint8_t motor1;                    /*!< first motor id */
    uint8_t counter1;                  /*!< first motor counter */
    uint8_t motor2;                    /*!< second motor id */
    uint16_t cmdid2;                   /*!< second motor cmdid */
    int16_t speed2;                    /*!< second motor signed speed */
    uint8_t counter2;                  /*!< second motor counter */
};

/** Start motor packet */
struct ftop_start_sync_motor {
    struct ftop_command_packet packet; /*!< common command header */
    uint8_t action;                    /*!< command action */
    uint8_t mask;                      /*!< group mask */
    uint16_t distance;                 /*!< distance */
    int16_t speed[FT_MAX_MOTORS];      /*!< signed speeds of motors */
    uint16_t cmdid[FT_MAX_MOTORS];     /*!< cmdids of motors */
};

/** magic type */
typedef uint32_t ftobject_magic;

/*!
 * Common start for all ftobjects
 */
typedef struct ftobject {
    ftobject_magic magic; /*!< ftobject magic value */
} ftobject;

/*!
 * Input value I1-I8
 */
typedef uint16_t ftvalue;

/*!
 * Client Inputs info
 */
typedef struct ftcvalues {
	uint32_t cmdids[FT_MAX_MOTORS];  /*!< M1-M4 cmdids */
} ftcvalues;

/*!
 * Inputs info
 */
typedef struct ftvalues {
	struct ftcvalues cmdvalues;         /*!< M1-M4 cmdids */
	uint16_t counters[FT_MAX_COUNTERS]; /*!< C1-C4 values */
    ftvalue inputs[FT_MAX_INPUTS];      /*!< I1-I8 values */
    uint8_t cstate[FT_MAX_COUNTERS];    /*!< C1-C4 state */
    uint16_t temperature;				/*!< temperature, unused */
    struct ft_faults ft_finfo;          /*!< faults/info values */
} ftvalues;

/** Values response packet */
struct ftop_response_values {
    struct ftop_command_packet packet; /*!< common response header */
    ftcvalues cvalues;                 /*!< values */
};

/** New values response packet */
struct ftop_nresponse_values {
    struct ftop_command_packet packet;	/*!< common response header */
    uint32_t nmask;						/*!< node mask */
    ftcvalues cvalues[FT_MAX_TXT];		/*!< ftcvalues from all slaves */
};

#endif /* _FT_IMPL_TYPES */
