/*
 Copyright (c) 2020 RTSoft. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
%module(docstring="This module is to control a new TXT4 controller") ft
%nothread;

%pythonbegin %{
import threading
from math import sqrt, log

def default_error_handler(message, exception):
  print(message)
  return False

def default_data_handler(ftTXT2):
  pass

def default_fault_handler(txt2, fts):
  pass

def getControllerInfoList(txt):
    """
    Get list of controllers
    :param txt: master ft.fttxt2
    :return: list of controllers info (master + slaves) 
    """
    list = []
    for i in range(txt.nslaves + 1):
        list.append(ftTXTControllerInfo(txt, i))
    return list

class ftTXT2exchange(threading.Thread):
    EB_INFO_UPDATE = 1
    EB_VALUES_UPDATE = 2
    EB_SLAVES_UPDATE = 4
    EB_FAULTS_UPDATE = 8

    def __init__(self, txt):
        threading.Thread.__init__(self)
        self.txt = txt

    def run(self):
        while True:
            events = self.txt.wait_event(self.EB_VALUES_UPDATE | self.EB_SLAVES_UPDATE | self.EB_FAULTS_UPDATE)
            if (events & self.EB_VALUES_UPDATE) != 0:
                self.txt.data_handler(self.txt)
            if (events & self.EB_SLAVES_UPDATE) != 0:
                self.txt.slave_handler(self.txt)
            if (events & self.EB_FAULTS_UPDATE) != 0:
                self.txt.fault_handler(self.txt, self.txt.get_finfo())

class ftTXTControllerInfo:
    """A class that represents master, slave TXT-controller information"""

    def __init__(self, txt, num):
        """
        :param txt: master ft.fttxt2
        :param num: int
        """
        self.name = txt.get_txt_name(num)
        self.role = self.name
        self.uid = txt.get_txt_uid(num)
        self.sn = txt.get_txt_SN(num)

    def __str__(self):
        return "(name=\"{}\", role=\"{}\", uid=0x{:08x})".format(self.name, self.role, self.uid)

    def get_role(self):
        """Get controller role from info"""
        return self.role

    def get_name(self):
        """Get controller name from info"""
        return self.name

%}

%{
#include "ftimpltypes.h"
#include "ftalloc.h"
#include "ftop.h"
#include "ftslave.h"

typedef struct {
    int values[8 + 4]; /* 8 inputs + 4 counters */
} Values;

%}

%feature("autodoc", "The class that represents the master/slave TXT4-controller") fttxt2;
typedef struct {
    %ignore magic;
    int magic;
} fttxt2;

%feature("autodoc", "The class that represents the device connected to the TXT4-controller") ftdevice;
typedef struct {
    %ignore magic;
    int magic;
} ftdevice;

%feature("pythonappend") fttxt2::fttxt2(const char *, int) %{
    self.txt_thread = ftTXT2exchange(txt=self)
    self.error_handler = default_error_handler
    self.data_handler = default_data_handler
    self.slave_handler = default_data_handler
    self.fault_handler = default_fault_handler
    self.getControllerInfoList = lambda: getControllerInfoList(self)
    self.getController = lambda info: self.get_slave_by_uid(info.uid)
    self.txt_thread.setDaemon(True)
    self.txt_thread.start()
%}

%feature("pythonappend") fttxt2::get_slave_by_uid(int) %{
    val.error_handler = default_error_handler
    val.data_handler = default_data_handler
    val.getControllerInfoList = lambda: []
    val.getController = None
%}

%feature("pythonappend") fttxt2::input(int) %{
    self.update_config()
%}

%feature("pythonappend") fttxt2::output(int) %{
    self.update_config()
%}

%feature("pythonappend") fttxt2::motor(int) %{
    self.update_config()
%}

%feature("pythonappend") fttxt2::resistor(int) %{
    self.update_config()
%}

%extend fttxt2 {
    %feature("autodoc", "TXT4 controller name") name;
    const char name[16];
    %feature("autodoc", "TXT4 controller board version") version;
    const int version;
    %feature("autodoc", "TXT4 controller firmware version") firmware;
    const int firmware;
    %feature("autodoc", "Number of slave controllers connected to the TXT4 controller") nslaves;
    const int nslaves;

%thread;
    fttxt2(const char *host = "auto", int port = 65000) {
        return ftop_alloc_txt(host, port);
    }
%nothread;

    ~fttxt2() {
        ftop_free_txt($self);
    }

	char *__str__() {
		static char temp[32];
		sprintf(temp, "txt2(\"%s\")", ftop_get_name($self));
		return temp;
	}

    %feature("autodoc", "Get master/slave TXT-controller role") get_role;
	const char *get_role() {
		return ftop_get_name($self); /* TODO: fix */
	}

	%feature("autodoc", "Get master/slave TXT-controller name") get_name;
	const char *get_name() {
		return ftop_get_name($self);
	}

    %pythoncode %{
        def coast(self, *motors):
            """Use coast mode to stop a group of motors
            :param motors: one or more motors
            """
            mask = 0
            for motor in motors:
                mask |= (1 << motor.motor)
            self.motor_coast(mask);
	%}

    %feature("autodoc", "Create resistor-object connected to I1-I8, num - 1-8") resistor;
    ftdevice *resistor(int num) {
        return ftop_alloc_resistor($self, num - 1 + ftdev_i1);
    }

    %feature("autodoc", "Create NTC-object connected to I1-I8, num - 1-8") ncttermistor;
    ftdevice *ncttermistor(int num) {
        return ftop_alloc_ncttermistor($self, num - 1 + ftdev_i1);
    }

    %feature("autodoc", "Create switch-object connected to I1-I8, num - 1-8") switch;
    ftdevice *switch(int num) {
        return ftop_alloc_switch($self, num - 1 + ftdev_i1);
    }

    %feature("autodoc", "Create switch-object connected to I1-I8, num - 1-8") input;
    ftdevice *input(int num) {
        return ftop_alloc_switch($self, num - 1 + ftdev_i1);
    }

    %feature("autodoc", "Create digital-object connected to I1-I8, num - 1-8") digital;
    ftdevice *digital(int num) {
        return ftop_alloc_digital($self, num - 1 + ftdev_i1);
    }

    %feature("autodoc", "Create voltmeter-object connected to I1-I8, num - 1-8") voltmeter;
    ftdevice *voltmeter(int num) {
        return ftop_alloc_voltmeter($self, num - 1 + ftdev_i1);
    }

    %feature("autodoc", "Create colorsensor-object connected to I1-I8, num - 1-8") colorsensor;
    ftdevice *colorsensor(int num) {
        return ftop_alloc_colorsensor($self, num - 1 + ftdev_i1);
    }

    %feature("autodoc", "Create trail_follower-object connected to I1-I8, num - 1-8") trail_follower;
    ftdevice *trail_follower(int num) {
        return ftop_alloc_trail_follower($self, num - 1 + ftdev_i1);
    }

    %feature("autodoc", "Create lamp-object connected to O1-O8, num - 1-8") lamp;
    ftdevice *lamp(int num) {
        return ftop_alloc_lamp($self, num - 1 + ftdev_o1);
    }

    %feature("autodoc", "Create lamp-object connected to O1-O8, num - 1-8") output;
    ftdevice *output(int num) {
        return ftop_alloc_lamp($self, num - 1 + ftdev_o1);
    }

    %feature("autodoc", "Create motor-object connected to M1-M4, num - 1-4") xsmotor;
    ftdevice *xsmotor(int num) {
        return ftop_alloc_motor($self, (num * 2) - 2 + ftdev_m1);
    }

    %feature("autodoc", "Create motor-object connected to M1-M4, num - 1-4") motor;
    ftdevice *motor(int num) {
        return ftop_alloc_encoder($self, (num * 2) - 2 + ftdev_m1);
    }

    %feature("autodoc", "Create encoder-object connected to M1-M4, num - 1-4") encoder;
    ftdevice *encoder(int num) {
        return ftop_alloc_encoder($self, (num * 2) - 2 + ftdev_m1);
    }

    %feature("autodoc", "Create counter-object connected to C1-C4, num - 1-4") counter;
    ftdevice *counter(int num) {
        return ftop_alloc_counter($self, num - 1 + ftdev_c1);
    }

    %feature("autodoc", "Create servo motor-object connected to S1-S3, num - 1-3") servo;
    ftdevice *servo(int num) {
        return ftop_alloc_servo($self, num - 1 + ftdev_s1);
    }

    %feature("autodoc", "Create ultrasonic-object connected to I1-I8, num - 1-8") ultrasonic;
    ftdevice *ultrasonic(int num) {
        return ftop_alloc_ultrasonic($self, num - 1 + ftdev_i1);
    }

    %feature("autodoc", "Check device code connected to I1-I8, num - 1-8") dicode;
    int dicode(int num) {
        return ftop_alloc_code($self, num - 1 + ftdev_i1);
    }

    %feature("autodoc", "Check device mode connected to I1-I8, num - 1-8") dimode;
    int dimode(int num) {
        return ftop_alloc_mode($self, num - 1 + ftdev_i1);
    }

%thread;
    %feature("autodoc", "Set FDCAN terminator mode, 0 - don't use terminator, 1 - use terminator") terminator;
    void terminator(int mode) {
        ftop_set_terminator($self, mode);
    }

    %feature("autodoc", "Set FDCAN bitrate") can_set_bitrate;
    void can_set_bitrate(int speed) {
        ftop_can_set_bitrate($self, speed);
    }

    %feature("autodoc", "Send config to the TXT2 controller") update_config;
    void update_config() {
        ftop_update_config($self);
    }

    %feature("autodoc", "Wait TXT2 controller event") wait_event;
    unsigned int wait_event(int mask) {
        unsigned int result = ftop_wait_event($self, mask);
        return result;
    }
%nothread;

    %feature("autodoc", "Get slave TXT2 controller name") get_txt_name;
    const char* get_txt_name(int num) {
        return ftop_get_txt_name($self, num);
    }

    %feature("autodoc", "Get slave TXT2 controller uid") get_txt_uid;
    unsigned int get_txt_uid(int num) {
        return ftop_get_txt_uid($self, num);
    }

    %feature("autodoc", "Get slave TXT2 controller serial number") get_txt_SN;
    const char*  get_txt_SN(int num) {
        return ftop_get_txt_SN($self, num);
    }

    %feature("autodoc", "Get slave TXT2 controller byuid") get_txt_SN;
    fttxt2*  get_slave_by_uid(int uid) {
        return ftop_get_slave_by_uid($self, uid);
    }

    %typemap(out) struct ft_faults {
       $result = PyDict_New();
       PyDict_SetItemString($result, "faults", PyInt_FromLong($1.faults));
       PyDict_SetItemString($result, "flags", PyInt_FromLong($1.flags));
       PyDict_SetItemString($result, "ocp", PyInt_FromLong($1.ocp));
       PyDict_SetItemString($result, "old", PyInt_FromLong($1.old));
    }

    %feature("autodoc", "Get TXT2 controller fault values") get_finfo;
    struct ft_faults get_finfo() {
        struct ft_faults fts;
        ftop_get_finfo($self, &fts);
        return fts;
    }

    %feature("autodoc", "Get string for typename defined by int") typename;
    const char *typename(int t) {
    	(void)$self;
        return ftmop_get_typename(t);
    }

    %typemap(out) Values {
        $result = PyList_New(12);
        for (int i = 0; i < 12; i++) {
            PyObject *o = PyInt_FromLong($1.values[i]);
            PyList_SetItem($result, i, o);
        }
    }

    %feature("autodoc", "Get TXT2 controller I1-I8 and C1-C4 raw values") get_values;
    Values get_values() {
        Values a;
        ftop_get_values($self, a.values);
        return a;
    }

    %feature("autodoc", "Use coast mode to stop a group of motors defined by mask 0-15") motor_coast;
    void motor_coast(int mask) {
		ftop_motor_coast($self, mask);
	}

    %feature("autodoc", "Close TXT connection") close;
    void close() {
	ftop_free_txt($self);
    }
};

%{
const char *fttxt2_name_get(fttxt2 *txt2) {
    return ftop_get_name(txt2);
}

unsigned int fttxt2_version_get(fttxt2 *txt2) {
    return ftop_get_version(txt2);
}

unsigned int fttxt2_firmware_get(fttxt2 *txt2) {
    return ftop_get_firmware(txt2);
}

unsigned int fttxt2_nslaves_get(fttxt2 *txt2) {
    return ftop_get_slaves_number(txt2);
}

%}

%extend ftdevice {
    %feature("autodoc", "Device connection") conn;
    const int conn;
    %feature("autodoc", "Motor number") motor;
    const int motor;
    %feature("autodoc", "Resistance for a resistor") resistance;
    const int resistance;
    %feature("autodoc", "Current value for a device") val;
    const int val;
    %feature("autodoc", "Current volatage for a device") voltage;
    const int voltage;
    %feature("autodoc", "Current temperature for a NTC device") temperature;
    const int temperature;
    %feature("autodoc", "Current distance for a ultrasonic device") distance;
    const int distance;
    %feature("autodoc", "Current distance for a encoder device") motor_distance;
    const int motor_distance;
    %feature("autodoc", "Current running state for a motor/encoder device") isrunning;
    const int isrunning;
    %feature("autodoc", "Current pwm value for a output device") pwm;
    const int pwm;
    ~ftdevice() {
        ftop_free_device($self);
    }

	char *__str__() {
		static char temp[128];
		void *txt = ftop_get_txt($self);
		int conn = ftop_get_conn($self);
		if (conn >= ftdev_i1 && conn <= ftdev_i8) {
			int code = ftop_alloc_code(txt, conn);
			int mode = ftop_alloc_mode(txt, conn);
			char *mmode;
			char *pmode = "";
			char *emode = "";
			if (mode & ftinp_ultrasonic) {
				mmode = "ultrasonic";
			} else {
				mmode = (mode & ftinp_analog) ? "analog" : "digital";
				pmode = (mode & ftinp_passive) ? "|passive" : "|active";
				emode = (mode & ftinp_r2) ? "|extended" : "";
			}
		    sprintf(temp, "ftdevice(%s, %d, %s%s%s)", ftmop_get_typename(code),
				conn - ftdev_i1 + 1, mmode, pmode, emode);
		} else if (conn >= ftdev_s1 && conn <= ftdev_s3) {
			sprintf(temp, "ftdevice(servo, %d)\n", conn - ftdev_s1 + 1);
		} else if (conn >= ftdev_c1 && conn <= ftdev_c4) {
			sprintf(temp, "ftdevice(counter, %d)\n", conn - ftdev_c1 + 1);
		} else if (conn >= ftdev_o1 && conn <= ftdev_o7) {
			int mode = ftop_alloc_mode(txt, conn);
			if (mode) {
				sprintf(temp, "ftdevice(motor, %d)", (conn - ftdev_o1) / 2 + 1);
			} else {
				sprintf(temp, "ftdevice(lamp, %d)", conn - ftdev_o1 + 1);
			}
		} else {
		    strcpy(temp,"ftdevice(unknown)");
		}
		return temp;
	}

    %pythoncode %{
        def state(self):
            """Get current state
            :return: 0-1 for digital input (I1-I8, C1-C4), voltage (mv) or distance
            """
            return self.val

        def setLevel(self, level):
            """Set pwm level
            :param level: new level 0-512
            """
            self.pwm_set(level)

        def setSpeed(self, speed):
            """Start the motor with selected speed
            :param speed: new speed -512 - 512"""
            self.start_speed(speed)

        def setDistance(self, distance, syncto = None):
            """Start the encoder (or pair of encoders) to travel a given distance"""
            self.start_distance(distance, syncto)

        def getCurrentDistance(self):
            """Get current encoder distance"""
            return self.motor_distance

        def finished(self):
            """Check that the encoder is not running and the previous operation is finished"""
            return not self.isrunning

        def ntcTemperature(self):
            """Get a temperature """
            # ftrobopy solution
            r = self.resistance
            if r != 0:
              x = log(r)
              y = x * x * 1.39323522
              z = x * -43.9417405
              T = y + z + 271.870481
            else:
              T = 10000
            return T

        def syncDistance(self, distance, *motors):
            """Start for a motor or a group of motors defined as ftdevice"""
            mask = 0
            if distance == 0:
                return;
            for motor in motors:
                mask |= (1 << motor.motor)
            self.start_sync_distance(distance, mask)

        def syncStart(self, *motors):
            """Start for a motor or a group of motors defined as ftdevice"""
            mask = 0
            for motor in motors:
                mask |= (1 << motor.motor)
            # 0 - unlimited distance
            self.start_sync_distance(0, mask)

        def syncStop(self, *motors):
            """Stop for a motor or a group of motors defined as ftdevice"""
            mask = 0
            for motor in motors:
                mask |= (1 << motor.motor)
            self.stop_sync(mask)

        def coast(self, *motors):
            """Use coast mode to stop a motor or a group of motors defined as ftdevice"""
            mask = 1 << self.motor
            for motor in motors:
                mask |= (1 << motor.motor)
            self.motor_coast(mask);
    %}
%thread;

    %feature("autodoc", "Use coast mode to stop a group of motors, defined by mask [0-15]") motor_coast;
	void motor_coast(int mask) {
	    void *txt = ftop_get_txt($self);
		ftop_motor_coast(txt, mask);
	}

	%feature("autodoc", "Set brightness [0-512] for a lamp") brightness_set;
    void brightness_set(int value) {
        ftop_set_brightness($self, value);
    }

    %feature("autodoc", "Set pwm value [0-512] for O1-O8 output, for example, lamp") pwm_set;
    void pwm_set(int value) {
        ftop_set_pwm($self, value);
    }

    %feature("autodoc", "Set a motor speed [-512, 512], for next distance operation") speed_set;
    void speed_set(int value) {
        ftop_motor_set_speed($self, value);
    }

    %feature("autodoc", "Start a motor with speed [-512, 512]") start_speed;
    void start_speed(int value) {
        ftop_motor_start_speed($self, value);
    }

    %feature("autodoc", "Start a motor (or pair of motors)") start_sync_distance;
    void start_distance(int distance, ftdevice *syncto = NULL) {
        ftop_motor_start_distance($self, distance, syncto);
    }

    %feature("autodoc", "Start a group of motors defined by mask [0, 15]") start_sync_distance;
    void start_sync_distance(int distance, int mask) {
        ftop_motor_start_sync_distance($self, distance, mask);
    }

    %feature("autodoc", "Stop a group of motors defined by mask [0, 15]") stop_sync;
    void stop_sync(int mask) {
        ftop_motor_stop_sync($self, mask);
    }

    %feature("autodoc", "Stop a motor") stop;
    void stop() {
        ftop_motor_stop($self);
    }

    %feature("autodoc", "Reset counter/encoder value") reset;
    void reset() {
        ftop_zero($self);
    }
%nothread;

    %feature("autodoc", "Set extended range for I1-I8, mode: 0 - normal, 1 - extended range") extend_range;
    void extend_range(int mode) {
        ftop_extend_range($self, mode);
    }

    %feature("autodoc", "Unconfigure an object and exclude from input cycle") unconfigure;
    void unconfigure() {
        ftop_free_device($self);
    }
};

%{
int ftdevice_resistance_get(ftdevice *dev) {
    return ftop_get_resistance(dev);
}
int ftdevice_voltage_get(ftdevice *dev) {
    return ftop_get_voltage(dev);
}
int ftdevice_temperature_get(ftdevice *dev) {
    return ftop_get_temperature(dev);
}
int ftdevice_val_get(ftdevice *dev) {
    return ftop_get_state(dev);
}
int ftdevice_pwm_get(ftdevice *dev) {
    return ftop_get_pwm(dev);
}
int ftdevice_distance_get(ftdevice *dev) {
    return ftop_get_distance(dev);
}
int ftdevice_motor_distance_get(ftdevice *dev) {
    return ftop_motor_get_distance(dev);
}
int ftdevice_isrunning_get(ftdevice *dev) {
    return ftop_motor_isrunning(dev);
}
int ftdevice_conn_get(ftdevice *dev) {
    return ftop_get_conn(dev);
}
int ftdevice_motor_get(ftdevice *dev) {
    return (ftop_get_conn(dev) - ftdev_m1) / 2;
}
%}
