/*
 * Copyright (c) 2020 RTSoft. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @file ftext.c
 * @brief Procedures to generate and process packages for slave controllers
 */

#include "ftext.h"
#include "ftosdep.h"

#include <string.h>
#include <FreeRTOS.h>
#include <cmsis_os.h>
#include <semphr.h>
#include <openamp_log.h>

/* Private variables */

static OS_MUTEX(ftext_mutex);

/* Private functions */

/**
 * @brief Update motor group mask for slave TXT2 controller
 * @param stxt external TXT2 controller
 * @param mask mask for motors in new group
 */
static void FixMotorMask(struct ft_shm_txt_remote *stxt, int mask)
{
	uint8_t mmask = stxt->op.mmask;
	uint8_t cmask = mask;

	if ((mask & -mask) == mask) {
		cmask = 0; // fix 1 bit combinations
	}

	if (mask & 1) {
		stxt->op.mmask = (mmask & (~mask << 4)) | cmask;
	} else {
		stxt->op.mmask = (mmask & ~mask & 0xf) | (cmask << 4);
	}
}

/**
 * @brief Get motors bit mask from all groups, where gmask & msk != 0
 * @param stxt external TXT2 controller
 * @param mask mask for motors in new group
 * @return motor mask
 */
static uint8_t GetTotalGroup(struct ft_shm_txt_remote *stxt, uint8_t mask)
{
	uint8_t nmask = mask;
	if (mask & stxt->lop.mmask) {
		nmask |= stxt->lop.mmask & 0xf;
	}
	if (mask & (stxt->lop.mmask >> 4)) {
		nmask |= (stxt->lop.mmask >> 4) & 0xf;
	}
	return nmask;
}

/**
 * @brief Destroy motor group, where gmask & msk != 0
 * @param stxt external TXT2 controller
 * @param mask mask for motors in new group
 */
static void DestroyGroup(struct ft_shm_txt_remote *stxt, uint8_t mask)
{
	uint8_t nmask = GetTotalGroup(stxt, mask);
	uint32_t ncmdid = 0;
	int i;

	for (i = 0; i < FT_MAX_MOTORS; i++) {
		if (nmask & (1 << i)) {
			int m = i * 2;
			if (stxt->op.pwm[m + 0] | stxt->op.pwm[m + 1]) {
				stxt->op.pwm[m + 0] = 0;
				stxt->op.pwm[m + 1] = 0;
			}
			if (stxt->common.mcommand[i].distance) {
				if (ncmdid == 0) {
					ncmdid = ++stxt->cmdid;
					if (!ncmdid)
						ncmdid = ++stxt->cmdid;
				}

				/* Stop motors in the same group */
				stxt->common.mcommand[i].distance = 0;
				stxt->common.mcommand[i].lcmdid = stxt->common.mcommand[i].ncmdid;
				stxt->common.mcommand[i].ncmdid = ncmdid;
				stxt->op.pwm[m + 0] = 0;
				stxt->op.pwm[m + 1] = 0;
			}
		}
	}
	stxt->op.mmask &= ~(nmask * 0x11);
}

/**
 * @brief Recalculate duty values (0-512) for remote motor
 * @param stxt external TXT2 controller
 * @param m motor 0,2,4 or 6
 * @param speed new speed [-512,512]
 */
static void SetSpeed(struct ft_shm_txt_remote *stxt, int m, int speed)
{
	if (speed >= 0) {
		if (speed > 512)
			speed = 512;
		stxt->op.pwm[m + 0] = speed;
		stxt->op.pwm[m + 1] = 0;
	} else {
		speed = -speed;
		if (speed > 512)
			speed = 512;
		stxt->op.pwm[m + 0] = 0;
		stxt->op.pwm[m + 1] = speed;
	}

	if (stxt->lop.opid == stxt->op.opid) {
		uint16_t nopid = stxt->lop.opid + 1;
		if (!nopid)
			nopid++;
		stxt->op.opid = nopid;
	}
}

/**
 * @brief Process sync start/stop command from A7 for remote motors
 * @param ftc A7 client
 * @param ftcmd client command
 * @param stxt external TXT2 controller
 */
static void StartSyncMotor(ftop_client *ftc, struct ftop_command_packet *ftcmd,
		struct ft_shm_txt_remote *stxt)
{
	int i;
	struct ftop_start_sync_motor *start = (void*) ftcmd;
	uint8_t mask = start->mask & 0xf;

	int dst = start->packet.dst; // Get slave controller number

	OS_LOCK(ftext_mutex);
    uint32_t ncmdid = stxt->cmdid + 1;

	if (!ncmdid)
		ncmdid++;

    stxt->cmdid = ncmdid;

	for (i = 0; i < FT_MAX_MOTORS; i++) {
		if (!(mask & (1 << i)))
			continue;

		int m = i * 2;

		stxt->common.counters[i] = 0;
		ftc->motor_cmdid[dst][i].mcmdid = ncmdid;
		ftc->motor_cmdid[dst][i].ncmdid = start->cmdid[i];
		stxt->common.mcommand[i].lcmdid = stxt->common.mcommand[i].ncmdid;
		stxt->common.mcommand[i].ncmdid = ncmdid;

		if (start->action == ftop_m_stop) {
			SetSpeed(stxt, m, 0);
			stxt->common.mcommand[i].distance = 0;
			stxt->op.distance[i] = 0;
		} else {
			SetSpeed(stxt, m, start->speed[i]);
			stxt->common.mcommand[i].distance = start->distance;
			stxt->op.distance[i] = start->distance;
			stxt->op.ccmdid[i] = ncmdid;
		}

		stxt->op.mcmdid[i] = ncmdid;
	}

	FixMotorMask(stxt, mask);

	OS_UNLOCK(ftext_mutex);
}

/**
 * @brief Process remote motor commands from A7
 * @param ftc A7 client
 * @param ftcmd client command
 * @param stxt external TXT2 controller
 */
static void ExtStartMotor(ftop_client *ftc, struct ftop_command_packet *ftcmd,
		struct ft_shm_txt_remote *stxt)
{
	struct ftop_start_motor *start = (void*)ftcmd;
    unsigned int m1 = start->motor1 - ftdev_m1; // even
    unsigned int m2 = start->motor2 - ftdev_m1; // even

	if (m1 >= FT_MAX_MOTORS * 2)
		return; /* Something wrong */

    unsigned int m12 = m1 / 2;
    unsigned int m22 = m2 / 2;
    uint32_t cmdid = stxt->cmdid + 1;

	if (!cmdid)
		cmdid++;

    stxt->cmdid = cmdid;

    OS_LOCK(ftext_mutex);

    switch(start->action) {
	case ftop_m_start:

		DestroyGroup(stxt, 1 << m12);
		SetSpeed(stxt, m1, start->speed1);
		break;

	case ftop_m_distance:

		DestroyGroup(stxt, 1 << m12);
		SetSpeed(stxt, m1, start->speed1);
		stxt->op.distance[m12] = start->distance;
		stxt->op.mcmdid[m12] = cmdid;
		stxt->op.ccmdid[m12] = cmdid;
		ftc->motor_cmdid[ftcmd->dst][m12].ncmdid = start->cmdid1;
		ftc->motor_cmdid[ftcmd->dst][m12].mcmdid = cmdid;
		break;

	case ftop_m_start_sync:

		if (m2 >= FT_MAX_MOTORS * 2) {
			return; /* Something wrong */
		}

		SetSpeed(stxt, m1, start->speed1);
		SetSpeed(stxt, m2, start->speed2);
		break;

	case ftop_m_distance_sync:

		if (m2 >= FT_MAX_MOTORS * 2) {
			return; /* Something wrong */
		}

		SetSpeed(stxt, m1, start->speed1);
		stxt->op.distance[m12] = start->distance;
		stxt->op.mcmdid[m12] = cmdid;
		stxt->op.ccmdid[m12] = cmdid;
		ftc->motor_cmdid[ftcmd->dst][m12].ncmdid = start->cmdid1;
		ftc->motor_cmdid[ftcmd->dst][m12].mcmdid = cmdid;

		SetSpeed(stxt, m2, start->speed2);
		stxt->op.distance[m22] = start->distance;
		stxt->op.mcmdid[m22] = cmdid;
		stxt->op.ccmdid[m22] = cmdid;
		ftc->motor_cmdid[ftcmd->dst][m22].ncmdid = start->cmdid2;
		ftc->motor_cmdid[ftcmd->dst][m22].mcmdid = cmdid;

		FixMotorMask(stxt, (1 << m12) | (1 << m22));
		break;

	case ftop_m_stop:

		DestroyGroup(stxt, 1 << m12);
		SetSpeed(stxt, m1, 0);
		stxt->op.distance[m12] = 0;
		stxt->op.mcmdid[m12] = cmdid;
		ftc->motor_cmdid[ftcmd->dst][m12].ncmdid = start->cmdid1;
		ftc->motor_cmdid[ftcmd->dst][m12].mcmdid = cmdid;
		break;
    }

	if (stxt->lop.opid == stxt->op.opid) {
		uint16_t nopid = stxt->lop.opid + 1;
		if (!nopid)
			nopid++;
		stxt->op.opid = nopid;
	}
	OS_UNLOCK(ftext_mutex);
}

/**
 * @brief Update input config for external TXT2 controller
 * @param [in] ftc A7 client
 * @param [in] stxt external TXT2 controller
 * @param [in] config new config
 * @param [in] dst controller id
 * @param [in] cusage component usage
 */
static void UpdateRemoteInputConfig(ftop_client *ftc, struct ft_shm_txt_remote *stxt, struct ft_config *config, uint32_t dst, uint32_t cusage)
{
	ftc->usage[dst] = cusage;
	uint32_t usage = CalculateTotalHWUsage(dst);

	for (int i = 0; i < FT_MAX_INPUTS; i++) {
		int mask = 1 << (i + ftdev_i1);

		if ((usage & mask) == 0) {
			stxt->current.config.inputs.inputs[i][0] = 0; /* go to unconfigured */
			stxt->current.config.inputs.inputs[i][1] = 0;
			continue;
		}

		if ((cusage & mask) == 0)
			continue;

		stxt->current.config.inputs.inputs[i][0] =
				config->inputs.inputs[i][0];
		stxt->current.config.inputs.inputs[i][1] =
				config->inputs.inputs[i][1];
	}

    stxt->current.usage = usage;
}

/**
 * @brief External engine initialization procedure
 */
void ftExtInit()
{
	ftext_mutex = xSemaphoreCreateMutex();
}

/* Exported functions */

/**
 * @brief Update local configuration for external slave TXT2 controller
 * @param [in] num external TXT2 controller number
 * @return pointer to the configuration copy
 */
struct ft_remote_config *ftExtUpdateConfig(int num)
{
	if (num < 1 || num >= FT_MAX_TXT)
		return NULL;

	struct ft_shm_txt_remote *s = (struct ft_shm_txt_remote *)getTXT(num);

	OS_LOCK(ftext_mutex);
	if (s->current.configid == s->last.configid) {
		OS_UNLOCK(ftext_mutex);
		return &s->last;
	}
	memcpy(&s->last, &s->current, sizeof(s->last));
	OS_UNLOCK(ftext_mutex);
	return &s->last;
}

/**
 * @brief Update config for external TXT2 controller
 * @param [in] ftc A7 client
 * @param [in] stxt external TXT2 controller
 * @param [in] config new config
 * @param [in] dst controller id
 * @param [in] cusage component usage
 */
void UpdateRemoteConfig(ftop_client *ftc, struct ft_shm_txt_remote *stxt,
		struct ft_config *config, uint32_t dst, uint32_t cusage)
{
	OS_LOCK(ftext_mutex);
	UpdateRemoteInputConfig(ftc, stxt, config, dst, cusage);
	uint32_t usage = stxt->current.usage;
	uint32_t gmask = 0;

	for (int i = 0; i < FT_MAX_MOTORS; i++) {
		int mask = 3 << (i * 2 + ftdev_m1);
		if (!(cusage & mask)) // use update for this client only
			continue;

		uint8_t u = config->motors.usage[i];
		stxt->current.config.motors.usage[i] = u;
		if (u == ft_output_mode) {
			uint8_t cmask = 1 << i;
			if (GetTotalGroup(stxt, cmask) != cmask)
				DestroyGroup(stxt, cmask);
		}
	}

	for (int i = 0; i < FT_MAX_MOTORS; i++) {
		int mask = 3 << (i * 2 + ftdev_m1);
		if (!(usage & mask)) // use update for all clients
			continue;

		uint8_t u = config->motors.usage[i];
		if (u == ft_motor_mode) {
			gmask |= 1 << i;
		}
	}

	stxt->op.usage = usage >> ftdev_o1;

	/* Check and fix group mask */
	if (stxt->op.mmask & gmask) {
		uint32_t g = stxt->op.mmask & 0xf;
		if ((g & gmask) != g) {
			DestroyGroup(stxt, g);
		}
	}
	if (stxt->op.mmask & (gmask << 4)) {
		uint32_t g = (stxt->op.mmask >> 4) & 0xf;
		if ((g & gmask) != g) {
			DestroyGroup(stxt, g);
		}
	}

	uint32_t configid = stxt->last.configid + 1;
	if (!configid)
		configid++;

	stxt->current.configid = configid;
	OS_UNLOCK(ftext_mutex);
}

/**
 * @brief Get copy of operation for remote slave, called by remote engine
 * @param stxt external TXT2 controller
 * @param op buffer for copy
 */
void getOp(struct ft_shm_txt_remote *stxt, struct ft_remote_op *op)
{
	OS_LOCK(ftext_mutex);
	memcpy(op, &stxt->op, sizeof(op[0]));
	memcpy(&stxt->lop, &stxt->op, sizeof(stxt->lop));
	OS_UNLOCK(ftext_mutex);
}

/**
 * @brief Process A7 command for external TXT2 controller
 * @param [in] ftc A7 client
 * @param [in] ftcmd client command
 * @param [in] stxt external TXT2 controller
 */
void ftProcessRequestExtRequest(ftop_client *ftc,
		struct ftop_command_packet *ftcmd, struct ft_shm_txt_remote *stxt)
{
	int dst = ftcmd->dst;
	if (dst >= FT_MAX_TXT)
		return;

	switch (ftcmd->cmd) {
	case ftop_op_ping:

		ftc->opened[dst] = 1;
		break;

	case ftop_op_get_config:

		break;

	case ftop_op_send_config:
		{
			struct ftop_setup_config *c = (void*)ftcmd;
			UpdateRemoteConfig(ftc, stxt, &c->config, dst, c->usage);
		}
		ftc->opened[dst] = 1;
		break;

	case ftop_op_start_motor:

		ExtStartMotor(ftc, ftcmd, stxt);
		break;

    case ftop_op_motor_coast:
		{
			struct ftop_set_mode *smode =  (void*)ftcmd;
			stxt->common.motormode = smode->mode;
		}
		break;

    case ftop_op_start_sync_motor:

		StartSyncMotor(ftc, ftcmd, stxt);
		break;

	case ftop_op_set_pwm:

		OS_LOCK(ftext_mutex);
		{
			struct ftop_set_pwm *c = (void*)ftcmd;
			int n;

			if ((c->devid >= ftdev_s1) && (c->devid <= ftdev_s3)) {
				n = c->devid - ftdev_s1 + FT_MAX_OUTPUTS;
			} else if (c->devid >= ftdev_o1 && c->devid <= ftdev_o8) {
				n = c->devid - ftdev_o1;
			} else {
				OS_UNLOCK(ftext_mutex);
				break;
			}
			stxt->op.pwm[n] = c->pwm;
			if (stxt->lop.opid == stxt->op.opid) {
				uint16_t nopid = stxt->lop.opid + 1;
				if (!nopid)
					nopid++;
				stxt->op.opid = nopid;
			}
		}
		OS_UNLOCK(ftext_mutex);
		break;

	case ftop_op_zero:
		OS_LOCK(ftext_mutex);
		{
			struct ftop_set_pwm *c = (void*)ftcmd;
			int i;
			if (c->devid >= ftdev_c1 && c->devid <= ftdev_c4) {
				i = c->devid - ftdev_c1;
			} else if (c->devid >= ftdev_m1 && c->devid <= ftdev_m4) {
				i = (c->devid - ftdev_m1) / 2;
			} else {
				OS_UNLOCK(ftext_mutex);
				break;
			}
			uint32_t cmdid = stxt->cmdid + 1;

			if (!cmdid)
				cmdid++;

			stxt->cmdid = cmdid;
			stxt->op.ccmdid[i] = cmdid;
			if (stxt->lop.opid == stxt->op.opid) {
				uint16_t nopid = stxt->lop.opid + 1;
				if (!nopid)
					nopid++;
				stxt->op.opid = nopid;
			}
		}
		OS_UNLOCK(ftext_mutex);
		break;

	case ftop_op_disconnect:
		break;

		/* Ignore !!! */
	case ftop_op_get_info:
		break;
	case ftop_op_set_mode:
		break;
	case ftop_op_terminator:
		break;
	case ftop_op_set_bitrate:
		break;
	case ftop_op_set_name:
		break;
	case ftop_op_power_down:
		break;
	case ftop_op_reset:
		/* TODO: */
		break;
	default: /* Something wrong, close connection */
		//log_info("Error cmd %d -> %d\n", ftcmd->cmd, ftcmd->dst);
		;
	}
}
