/*
 * Copyright (c) 2020 RTSoft. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @file ftserver.c
 * @brief Command processor for 'firmware"
 */

#include "ftconfig.h"
#include "ftext.h" /* includes ftserver.h */
#include "ftosdep.h"
#include "main.h"
#include "DRV8908.h"
#include "HWLoop.h"
#include "HWFDCAN.h"
#include "HWServo.h"
#include "ftcalib.h"
#include "ftservermem.h"

#include <openamp_log.h>
#include <string.h>
#include <FreeRTOS.h>
#include <cmsis_os.h>
#include <semphr.h>

/* Fault/alert pins */
static uint32_t pwrAlertPin; /*!< Power (+9.25V) alert pin value */
uint32_t drvAlertPin; /*!< DRV8908 alert pin value */
static uint32_t v5AlertPin;  /*!< Servo +5V alert pin value */

/* Clients info */
extern uint32_t master_usage;
ftop_client clients[FT_MAX_CLIENTS]; /*!< information about local clients connection */

extern IPCC_HandleTypeDef hipcc; /*!< IPCC HAL handle */
extern int nslaves;				 /*!< Number of slaves at this moment */

static const char alnum[] = {
		"ABCDEFGHIJKLMNOPQRSTUVWXYZ"
		"abcdefghijklmnopqrstuvwxyz"
		"0123456789"
};

extern void HWCountersUpdate(uint32_t *counters, uint8_t *cstate);
void PowerDown();

static uint8_t GetTotalGroup(uint8_t mask);

/*
 * Constants
 */
static struct ft_config zero_config; /* Unconfigured */

static OS_MUTEX(m4_mutex);
OS_MUTEX(out_mutex);
OS_MUTEX(drv8908_mutex);

/*
 * Common information
 */
static struct ft_config m4_config;
static struct ft_config values_config;
static unsigned int current_config_id;
static unsigned int m4_config_id;
static unsigned int values_config_id;
static int mode = 0; /* single */
static uint32_t m4_valuesid = 0;

static ftvalues m4_values;

/* Main structure initialization */
struct ft_shm_server txt2_server = {
	{"FTSH", FT_MAX_TXT, sizeof(txt2_server), &txt2_server.uids[0],
		&txt2_server.txts[0], &txt2_server.stat},};

struct ft_shm_stat *ftstat = &txt2_server.stat;
uint32_t *task_times = &txt2_server.task_times[0]; /*!< Common information about tasks time */

/* Response with information about a board */
static struct ftop_response_txt_info info = {
    {sizeof(info), ftop_op_get_info, 0, 0},
	.vaddr = (uint32_t)&txt2_server.current.values,
	.taddr = (uint32_t)&txt2_server.current,
	.uaddr = (uint32_t)&txt2_server,
};

/* Response for get configure request */
static struct ftop_response_config rconfig = {
    {sizeof(rconfig), ftop_op_get_config, 0, 0},
};

static uint32_t ops = 0;
static uint32_t zops = 0;

/**
 * @brief Get shared memory structure for the slave controller
 * @param num controller number
 * @return shared memory structure pointer for the slave controller
 */
struct ft_shm_txt *getTXT(int num)
{
	return txt2_server.txts[num];
}

/**
 * @brief Check uid/suid/cid combination
 * @param uid master controller uid
 * @param suid slave controller uid
 * @param cid CAN specific number from slave controller
 * @return shared memory structure pointer for the slave controller
 */
struct ft_shm_txt *Ext_TXT2_Check(ftuid uid, ftuid suid, int cid)
{
	uint32_t luid = buid;

	if (cid <= 0 || cid >= 8)
		return NULL;

	if (luid != uid)
			return NULL;

	struct ft_shm_txt *txt = txt2_server.txts[cid];
	if (txt->common.uid == suid)
		return txt;

	return NULL;
}

/**
 * @brief Fix C1-C4 and alarm values based on GPIO inputs, called for each ms
 */
void FTCountersUpdate()
{
	txt2_server.stat.runtime = HAL_GetTick();
    HWCountersUpdate(txt2_server.counters, txt2_server.cstate);

	for (int i = 0; i < FT_MAX_COUNTERS; i++) {
		txt2_server.current.values.counters[i] = txt2_server.counters[i]
				- txt2_server.fcounters[i];
	}
	*(uint32_t*)txt2_server.current.values.cstate = *(uint32_t*)txt2_server.cstate;

	/* Check ALERT pins */
	pwrAlertPin = (PWR_BUCK_ALERT_GPIO_Port->IDR >> NPIN(PWR_BUCK_ALERT_Pin)) & 1;
	if (!pwrAlertPin)
		txt2_server.stat.pwralert++;

	drvAlertPin = !((MOTOR_NFAULT_GPIO_Port->IDR >> NPIN(MOTOR_NFAULT_Pin)) & 1);
	if (drvAlertPin)
		txt2_server.stat.drvalert++;

#if TXT2_BOARD_REVISION == 1
	v5AlertPin = (SERVO_OC_GPIO_Port->IDR >> NPIN(SERVO_OC_Pin)) & 1;
	if (v5AlertPin)
		txt2_server.stat.servoalert++;
#endif
}

/**
 * @brief Procedure for consistant copy of values
 */
static void CopyM4Values()
{
	struct ftvalues *fc = &txt2_server.current.values;
	int i;

	for (i = 0; i < FT_MAX_MOTORS; i++) {
		fc->cmdvalues.cmdids[i] = m4_values.cmdvalues.cmdids[i];
	}
	for (i = 0; i < FT_MAX_INPUTS; i++) {
		fc->inputs[i] = m4_values.inputs[i];
	}

	fc->ft_finfo = m4_values.ft_finfo;
}

/**
 * @brief Input process thread, must not communicate
 * @param info unused
 */
void ftopThread(void const *info)
{
    struct ft_shm_txt *current = &txt2_server.current;
    (void)info;
    uint32_t tloop = 0;

    while(1) {
#if TXT2_BOARD_REVISION == 1
        uint32_t ctime = TIM5->CNT; /* clock * 10 */
#else
		uint32_t ctime = HAL_GetTick();
#endif
		OS_LOCK(m4_mutex);

		/* Copy current config */
		memcpy(&values_config, &m4_config, sizeof(values_config));
		values_config_id = m4_config_id;
		memcpy(&m4_config, &current->config, sizeof(m4_config));
		m4_config_id = current_config_id;

		/* Copy current values */
		CopyM4Values();
		m4_valuesid++;

		OS_UNLOCK(m4_mutex);

		/* Start new cycle for I1-I8 inputs */
		HWLoop(&m4_values, &m4_config.inputs);
		DRV8908_Check_Faults(&m4_values.ft_finfo);
		m4_values.ft_finfo.faults = pwrAlertPin | (drvAlertPin << 1) | (v5AlertPin << 2);
#if TXT2_BOARD_REVISION == 1
		uint32_t eloop = TIM5->CNT;
		uint32_t ediff = eloop - ctime;
		if (ediff > tloop) {
			tloop = ediff;
			txt2_server.stat.maxhwlooptime = tloop; /* 0.01 ms */
		}
		if (ediff < 100) { /* 1 ms */
			osDelay(1); /* Give a chance */
		}
#else
		uint32_t eloop = HAL_GetTick();
		if ((eloop - ctime) > tloop) {
			tloop = eloop - ctime;
			txt2_server.stat.maxhwlooptime = tloop * 100;
		}
		if (eloop == ctime) {
			osDelay(1); /* Give a chance */
		}
#endif /* TXT2_BOARD_REVISION */
    }
}

/**
 * @brief Procedure to generate a name for discovery
 */
static void GenerateShortName()
{
	int sz = sizeof(alnum) - 1; /* Must be 62 */
	uint32_t uid = buid;

	memset(txt2_server.current.common.device_name, 0,
			sizeof(txt2_server.current.common.device_name));
#if 1 /* 4 characters name, can produce same name for different controllers */
	uid %= sz * sz * sz * sz - 1;
	txt2_server.current.common.device_name[0] = alnum[uid % sz];
	uid /= sz;
	txt2_server.current.common.device_name[1] = alnum[uid % sz];
	uid /= sz;
	txt2_server.current.common.device_name[2] = alnum[uid % sz];
	uid /= sz;
	txt2_server.current.common.device_name[3] = alnum[uid % sz];
#else /* Up to 6 characters name, unique for controller */
	int i;
	if ((uid & 0xff000000) == 0) { /* 3-5 */
		if ((uid & 0xff00) == 0) { /* 3 */
			uid = ((uid >> 16) & 0xff00) | (uid & 0xff);
			txt2_server.current.common.device_name[2] = alnum[uid % sz];
			uid /= sz;
			txt2_server.current.common.device_name[1] = alnum[uid % sz];
			uid /= sz;
			txt2_server.current.common.device_name[0] = 'f' + uid;
		} else { /* 5 */
			for (i = 4;i > 0; i--) {
				txt2_server.current.common.device_name[i] = alnum[uid % sz];
				uid /= sz;
			}
			txt2_server.current.common.device_name[0] = 'x' + uid;
		}

	} else { /* Full case 6 */
		for (i = 5; i > 0; i--) {
			txt2_server.current.common.device_name[i] = alnum[uid % sz];
			uid /= sz;
		}
		txt2_server.current.common.device_name[0] = 'a' + uid;
	}
#endif
}

/**
 * @brief Initialization procedure
 */
void ftInitServer()
{
    /* Init common structures */
    m4_mutex = xSemaphoreCreateMutex();
    out_mutex = xSemaphoreCreateMutex();
    drv8908_mutex = xSemaphoreCreateMutex();

	/* Initialization for common structures and shared memory */
    txt2_server.uids[0] = buid;
    txt2_server.current.common.uid = txt2_server.uids[0];
    txt2_server.txts[0] = &txt2_server.current;
	for (int i = 0; i < FT_MAX_SLAVES; i++) {
		txt2_server.txts[i + 1] = &txt2_server.slaves[i].common;
	}

	/* Generate short name based on UID */
    GenerateShortName();

    /* Setup versions */
	txt2_server.current.common.version = FT_M4_VERSION;
	txt2_server.current.common.firmware = FT_M4_FIRMWARE;
	sprintf(txt2_server.current.common.sn, "ft%8.8lX", txt2_server.uids[0]);

    values_config = txt2_server.current.config;
    current_config_id = 1;

    /* Fix header */
    txt2_server.header.uids = &txt2_server.uids[0];
    txt2_server.header.txts = txt2_server.txts;
    txt2_server.header.task_times = task_times;

	/* Initialization for clients */
	for(int i = 0; i < FT_MAX_CLIENTS; i++) {
	    ftop_client *c = &clients[i];
	    c->id = i;
	    c->socket = i + 2;
	    memset(c->opened, 0, sizeof(c->opened));
	}
}

/**
 * @brief Send TXT2 controller info to A7 client
 * @param ftc A7 client
 */
static void SendInfo(ftop_client *ftc)
{
    int cmdid = 0;

    info.packet.dst = 0;
    info.packet.cmdid = cmdid;
    info.mode = txt2_server.current.common.mode;
    info.uid = txt2_server.current.common.uid;
    info.slaves = nslaves;
    info.version = txt2_server.current.common.version;
    info.firmware = txt2_server.current.common.firmware;

    strncpy(info.name, txt2_server.current.common.device_name, sizeof(info.name));

    ftserver_send_response(ftc, &info, sizeof(info));
}

/**
 * @brief Prepare TXT2 controller values for master
 * @param values buffer for values
 */
void getValues(struct ft_remote_values *values)
{
	struct ft_shm_txt *current = &txt2_server.current;
	int i;

	OS_LOCK(m4_mutex);
	memcpy(&values->values, &current->values, sizeof(values->values));
	values->valuesid = m4_valuesid;
	OS_UNLOCK(m4_mutex);

	OS_LOCK(out_mutex);
	struct master_motor *mm = txt2_server.mmotor_cmdid;
	uint32_t *fcv = values->values.cmdvalues.cmdids;
	for (i = 0; i < FT_MAX_MOTORS; mm++, i++) {
		if (mm->lcmdid != mm->ncmdid) {
			int32_t diff = current->mcommand[i].lcmdid
					- txt2_server.mmotor_cmdid[i].mcmdid;
			if (diff >= 0) {
				mm->lcmdid = mm->ncmdid;
			}
		}
		*fcv++ = mm->lcmdid;
	}
	OS_UNLOCK(out_mutex);
}

/**
 * @brief Send TXT2 controller config to A7 client
 * @param ftc A7 client
 * @param ftcmd A7 client command
 * @param dst TXT2 controller number
 */
static void SendConfig(ftop_client *ftc, struct ftop_command_packet *ftcmd, int dst)
{
	int cmdid = ftcmd == NULL ? 0 : ftcmd->cmdid;

	OS_LOCK(m4_mutex);

	ftc->configid = values_config_id;
	rconfig.packet.dst = dst;
	rconfig.packet.cmdid = cmdid;
	rconfig.configid = values_config_id;
	rconfig.config = values_config;
	OS_UNLOCK(m4_mutex);

	/* Send copy */
	ftserver_send_response(ftc, &rconfig, sizeof(rconfig));
}

/**
 * @brief Set motor duty values for local TXT2 controller
 * @param m motor 0, 2, 4 or 6
 * @param speed new speed [-512, 512]
 */
static void SetSpeed(int m, int speed)
{
	struct ft_shm_txt *current = &txt2_server.current;
	if (speed > 0) {
		if (speed > 512)
			speed = 512;
		current->opwm[m + 0] = speed;
		current->opwm[m + 1] = 0;
	} else {
		speed = -speed;
		if (speed > 512)
			speed = 512;
		current->opwm[m + 0] = 0;
		current->opwm[m + 1] = speed;
	}
}

/**
 * @brief Destory motor group for local TXT2 controller
 * @param mask new motors mask
 */
static void DestroyGroup(uint8_t mask)
{
	/* Destroy motor group and stop motor */
	struct ft_shm_txt *current = &txt2_server.current;
	uint32_t ncmdid = 0;
	uint8_t nmask = GetTotalGroup(mask);
	int i;

	for (i = 0; i < FT_MAX_MOTORS; i++) {
		if (nmask & (1 << i)) {
			int m = i * 2;
			if (current->opwm[m + 0] | current->opwm[m + 1]) {
				current->opwm[m + 0] = 0;
				current->opwm[m + 1] = 0;
			}
			if (current->mcommand[i].distance) {
				if (ncmdid == 0) {
					ncmdid = ++txt2_server.ncmdid;
					if (!ncmdid)
						ncmdid = ++txt2_server.ncmdid;
				}
				current->mcommand[i].distance = 0;
				current->mcommand[i].lcmdid = current->mcommand[i].ncmdid;
				current->mcommand[i].ncmdid = ncmdid;
				txt2_server.break_info[i].reached = 0;
			}
		}
	}
	txt2_server.mmask &= ~(nmask * 0x11);
}

/**
 * @brief Get motor group mask for local TXT2 controller
 * @param mask new motors mask
 * @return Total group mask
 */
static uint8_t GetTotalGroup(uint8_t mask)
{
	uint8_t nmask = mask & 0xf;
	if (mask & txt2_server.mmask) {
		nmask |= txt2_server.mmask & 0xf;
	}
	if (mask & (txt2_server.mmask >> 4)) {
		nmask |= (txt2_server.mmask >> 4) & 0xf;
	}
	return nmask;
}

/**
 * @brief Common local procedure to start/stop group of motors
 * @param ftc A7 client
 * @param distance new distance >= 0
 * @param cmdids pointer to cmdids
 * @param speed pointer to speeds [-512, 512]
 * @param mask mask for active motors
 * @param tmask mask for all motors in old groups
 */
static void CommonMotorsStart(ftop_client *ftc, uint16_t distance,
		uint16_t *cmdids, int16_t *speed, uint8_t mask, uint8_t tmask)
{
	struct ft_shm_txt *current = &txt2_server.current;
	int i;
    uint32_t ncmdid = ++txt2_server.ncmdid;
	if (!ncmdid)
		ncmdid = ++txt2_server.ncmdid;

	for (i = 0; i < FT_MAX_MOTORS; i++) {
		if (!(tmask & (1 << i)))
			continue;

		int m = i * 2;

		current->counters[i] = 0;
		ftc->motor_cmdid[0][i].mcmdid = ncmdid;
		ftc->motor_cmdid[0][i].ncmdid = cmdids[i];
		current->mcommand[i].lcmdid = current->mcommand[i].ncmdid;
		current->mcommand[i].ncmdid = ncmdid;

		txt2_server.break_info[i].reached = 0;
		if ((mask & (1 << i)) == 0) {
			SetSpeed(m, 0);
			current->mcommand[i].distance = 0;
			txt2_server.break_info[i].reached = 0;
		} else {
			SetSpeed(m, speed[i]);
			if (speed[i])
				txt2_server.fcounters[i] = txt2_server.counters[i];
			current->mcommand[i].distance = distance;
			txt2_server.break_info[i].breakmode = 0;
			txt2_server.break_info[i].lcount = 0;
			txt2_server.break_info[i].ltime = 0;
			txt2_server.break_info[i].dtime = 0;
		}
	}

	/* Fix common group mask */
	uint8_t nmask = txt2_server.mmask & ~(tmask * 0x11);
	if ((mask & -mask) != mask) { /* one bit or zero */
		nmask = (mask & 1) ? ((nmask & 0xf0) | mask) : ((nmask & 0xf) | (mask << 4));
	}
	txt2_server.mmask = nmask;
}

/**
 * @brief Process sync start/stop commands
 * @param ftc A7 client
 * @param ftcmd new command
 */
static void StartSyncMotor(ftop_client *ftc, struct ftop_command_packet *ftcmd)
{
	struct ftop_start_sync_motor *start = (void*) ftcmd;
	uint8_t mask = start->mask;
	uint8_t tmask = GetTotalGroup(mask);

	if (start->action == ftop_m_stop) {
		mask = 0;
	}
	txt2_server.current.motormode &= ~mask;

	CommonMotorsStart(ftc, start->distance, start->cmdid, start->speed, mask, tmask);
}

/**
 * @brief Process motor commands
 * @param ftc A7 client
 * @param ftcmd new command
 */
static void StartMotor(ftop_client *ftc, struct ftop_command_packet *ftcmd)
{
	static uint32_t mtime = 0;
    uint32_t stime = HAL_GetTick();
    struct ftop_start_motor *start = (void*)ftcmd;
    struct ft_shm_txt *current = &txt2_server.current;
    unsigned int m1 = start->motor1 - ftdev_m1; // even
    unsigned int m2 = start->motor2 - ftdev_m1; // even

	if (m1 >= FT_MAX_MOTORS * 2)
		return; /* Something wrong */

    unsigned int m12 = m1 / 2;
    unsigned int m22 = m2 / 2;

    current->motormode &= ~(1 << m12);

    switch(start->action) {

    case ftop_m_start:

        SetSpeed(m1, start->speed1);
        OS_LOCK(drv8908_mutex);
        vTaskSuspendAll ();
        DRV8908_motor_speed(m1, &current->opwm[m1]);
        xTaskResumeAll ();
        OS_UNLOCK(drv8908_mutex);
        break;

    case ftop_m_distance:

		{
			int16_t speed[4] = {0};
			uint16_t cmdids[4] = {0};
			uint8_t mask = 1 << m12;
			uint8_t tmask = GetTotalGroup(mask);
			speed[m12] = start->speed1;
			cmdids[m12] = start->cmdid1;
			CommonMotorsStart(ftc, start->distance, cmdids, speed, mask, tmask);
		}
        break;

    case ftop_m_start_sync:

		if (m2 >= FT_MAX_MOTORS * 2)
			return; /* Something wrong */

		current->motormode &= ~(1 << m22);
        SetSpeed(m1,  start->speed1);
        SetSpeed(m2,  start->speed2);
        OS_LOCK(drv8908_mutex);
        vTaskSuspendAll ();
        DRV8908_motor_speed(m1, &current->opwm[m1]);
        DRV8908_motor_speed(m2, &current->opwm[m2]);
        xTaskResumeAll ();
        OS_UNLOCK(drv8908_mutex);
        break;

    case ftop_m_distance_sync:

		if (m2 >= FT_MAX_MOTORS * 2)
			return; /* Something wrong */

		current->motormode &= ~(1 << m22);
		{
			int16_t speed[4] = { 0 };
			uint16_t cmdids[4] = { 0 };
			uint8_t mask = (1 << m12) | (1 << m22);
			uint8_t tmask = GetTotalGroup(mask);
			speed[m12] = start->speed1;
			cmdids[m12] = start->cmdid1;
			speed[m22] = start->speed2;
			cmdids[m22] = start->cmdid1;
			CommonMotorsStart(ftc, start->distance, cmdids, speed, mask, tmask);
		}
        break;

    case ftop_m_stop:

		{
			int16_t speed[4] = {0};
			uint16_t cmdids[4] = {0};
			uint8_t mask = 1 << m12;
			uint8_t tmask = GetTotalGroup(mask);
			cmdids[m12] = start->cmdid1;
			CommonMotorsStart(ftc, start->distance, cmdids, speed, 0, tmask);
		}
        break;
    }

    uint32_t etime = HAL_GetTick();
    etime -= stime;
	if (etime > mtime) {
		mtime = etime;
	}
}

/**
 * @brief Process control command from master
 * @param stxt remote master structure
 * @param op new operation
 */
void ProcessRemoteOp(struct ft_master_remote *stxt, struct ft_remote_op *op)
{
	if (op->opid == stxt->lop.opid)
		return;

	OS_LOCK(out_mutex);

	struct ft_shm_txt *current = &txt2_server.current;
	int i;

	/* Check all Motors */
    uint32_t ncmdid = ++txt2_server.ncmdid;
	if (!ncmdid)
		ncmdid = ++txt2_server.ncmdid;

	uint32_t gmask = 0;

	for (i = 0; i < FT_MAX_MOTORS; i++) {
		if (!(op->usage & (3 << (2 * i))))
			continue;

		if (txt2_server.current.config.motors.usage[i] == ft_output_mode)
			continue;

		gmask |= 1 << i;
	}

    /* Mix group bitmask */
    uint8_t omask = op->mmask & 0x0f;
    uint8_t tmask = txt2_server.mmask;
    uint8_t cmask = tmask & 0x0f;
    uint8_t ggmask = gmask * 0x11;
    uint8_t mmask = 0;

	if (omask) {
		if (omask != cmask)
			DestroyGroup(cmask);
		mmask |= omask;
	} else if (cmask & ggmask) {
		DestroyGroup(cmask);
	} else {
		mmask |= cmask;
	}

	omask = op->mmask & 0x0f0;
	cmask = txt2_server.mmask & 0x0f0;

	if (omask) {
		if (omask != cmask)
			DestroyGroup(cmask >> 4);
		mmask |= omask;
	} else if (cmask & ggmask) {
		DestroyGroup(cmask >> 4);
	} else {
		mmask |= cmask;
	}

    txt2_server.mmask = mmask;

	for (i = 0; i < FT_MAX_MOTORS; i++) {
		if (!(op->usage & (3 << (2 * i))))
			continue;

		if (op->mcmdid[i] != stxt->lop.mcmdid[i]) {
			/* New motor operation */
			current->mcommand[i].distance = op->distance[i];
			current->mcommand[i].lcmdid = current->mcommand[i].ncmdid;
			current->mcommand[i].ncmdid = ncmdid;
			txt2_server.mmotor_cmdid[i].lcmdid = txt2_server.mmotor_cmdid[i].ncmdid;
			txt2_server.mmotor_cmdid[i].ncmdid = op->mcmdid[i];
			txt2_server.mmotor_cmdid[i].mcmdid = ncmdid;
			txt2_server.break_info[i].breakmode = 0;
			txt2_server.break_info[i].dtime = 0;
			txt2_server.break_info[i].lcount = 0;
			txt2_server.break_info[i].ltime = 0;
			txt2_server.break_info[i].reached = 0;
		}
	}

	/* Check all Counters */
	for (i = 0; i < FT_MAX_COUNTERS; i++) {
		if (!(op->usage & (1 << (i + FT_MAX_OUTPUTS))))
			continue;

		if (op->ccmdid[i] != stxt->lop.ccmdid[i]) {
			/* New counter operation */
			txt2_server.fcounters[i] = txt2_server.counters[i];
		}
	}

	/* Check all PWMs */
	for (i = 0; i < FT_MAX_OUTPUTS; i++) {
		if (!(op->usage & (1 << i)))
			continue;
		current->opwm[i] = op->pwm[i];
	}

	/* Check all Servo */
	for(i = 0; i < FT_MAX_SERVO; i++) {
		if (!(op->usage & (1 << (i + FT_MAX_OUTPUTS + FT_MAX_COUNTERS))))
			continue;

		uint16_t ipwm = op->pwm[i + FT_MAX_OUTPUTS];
		uint16_t cpwm = ipwm >= 512 ? 512 : ipwm;
		HWServo_SetDuty (i, cpwm);
	}

	OS_UNLOCK(out_mutex);

	memcpy(&stxt->lop, op, sizeof(stxt->lop));
}

/**
 * @brief Process new config from local/remote source
 * @param cusage resources usage value for config
 * @param c new config
 */
void ProcessNewConfig(uint32_t cusage, struct ft_config *c)
{
	OS_LOCK(m4_mutex);

    uint32_t usage = CalculateTotalHWUsage(0);
    int i;

	for (i = 0; i < FT_MAX_INPUTS; i++) {
		int mask = 1 << (i + ftdev_i1);

		if ((usage & mask) == 0) {
			txt2_server.current.config.inputs.inputs[i][0] = 0; /* go to unconfigured */
			txt2_server.current.config.inputs.inputs[i][1] = 0;
			continue;
		}

		if ((cusage & mask) == 0)
			continue;

		txt2_server.current.config.inputs.inputs[i][0] = c->inputs.inputs[i][0];
		txt2_server.current.config.inputs.inputs[i][1] = c->inputs.inputs[i][1];
	}

	current_config_id++;

	OS_UNLOCK(m4_mutex);

	OS_LOCK(out_mutex);

	/* Process motor config */
	uint32_t gmask = 0;
	for (i = 0; i < FT_MAX_MOTORS; i++) {
		int mask = 3 << (i * 2 + ftdev_m1);
		if (!(usage & mask))
			continue;

		txt2_server.current.config.motors.usage[i] = c->motors.usage[i];
		if (c->motors.usage[i] == ft_motor_mode) {
			gmask |= 1 << i;
		}
	}

	if (gmask) { /* op->mmask may be zero ! */
		uint8_t mmask = txt2_server.mmask;
		uint32_t g = mmask & 0xf;

		if ((g & gmask) && (g & gmask) != g) {
			DestroyGroup(g);
			mmask &= 0xf0;
		}
		g = (mmask >> 4) & 0xf;
		if ((g & gmask) && (g & gmask) != g) {
			DestroyGroup(g);
			mmask &= 0x0f;
		}
		txt2_server.mmask = mmask;
	}

	/* Stop unused resources */
	for (i = 0; i < FT_MAX_OUTPUTS; i++) {
		if (!(usage & (1 << (i + ftdev_o1)))) {
			txt2_server.current.opwm[i] = 0;
		}
	}
	OS_UNLOCK(out_mutex);
}

/**
 * @brief Process new command for remote TXT2 controller
 * @param ftc A7 client
 * @param buffer buffer with command
 */
void ftProcessRequest(ftop_client *ftc, unsigned char *buffer)
{
    struct ftop_command_packet *ftcmd = (void*)buffer;
    ops++;

    txt2_server.stat.clients |= 1 << ftc->id;
    txt2_server.stat.commands++;

    /* Is it remote request ? */
 	if (ftcmd->dst) {
		ftProcessRequestExtRequest(ftc, ftcmd,
				&txt2_server.slaves[ftcmd->dst - 1]);
		return;
	}

    switch(ftcmd->cmd) {

    case ftop_op_ping:

		ops--; /* Skip ping requests */
		ftc->opened[0] = 1;
		break;

    case ftop_op_nping:

		ops--; /* Skip ping requests */
		{
			struct ftop_set_mode *smode = (void*) buffer;
			uint32_t mask = smode->mode;

			for (int i = 0; i < FT_MAX_TXT; i++) {
				if (mask & (1 << i))
					ftc->opened[i] = 1;
			}
		}
		break;

    case ftop_op_set_mode:

        {
			struct ftop_set_mode *smode = (void*) buffer;

			mode = smode->mode;
			txt2_server.current.common.mode = mode;
			if (mode) {
				HWFDCAN_set_mode(mode);
			}
        }
        SendInfo(ftc);
        break;

    case ftop_op_get_info: /* Send txt info back */

		SendInfo(ftc);
		ftc->opened[0] = 1;
		break;

    case ftop_op_get_config: /* Send config back */

        SendConfig(ftc, ftcmd, 0);
        break;

    case ftop_op_send_config: /* Process new config */

        {
            struct ftop_setup_config *c = (void*)buffer;
            ftc->usage[0] = c->usage;

            ProcessNewConfig(c->usage, &c->config);
        }
        break;

    case ftop_op_start_motor:

	    OS_LOCK(out_mutex);
        StartMotor(ftc, ftcmd);
        OS_UNLOCK(out_mutex);
        break;

    case ftop_op_start_sync_motor:

		OS_LOCK(out_mutex);
		StartSyncMotor(ftc, ftcmd);
		OS_UNLOCK(out_mutex);
		break;

    case ftop_op_motor_coast:
		{
			struct ftop_set_mode *smode =  (void*)buffer;
			OS_LOCK(out_mutex);
			txt2_server.current.motormode = smode->mode;
			for (int i = 0; i < FT_MAX_MOTORS; i++) {
				if (smode->mode & (1 << i)) {
					txt2_server.current.opwm[2 * i + 0] = 0;
					txt2_server.current.opwm[2 * i + 1] = 0;
				}
			}
			OS_UNLOCK(out_mutex);
		}
		break;

    case ftop_op_set_pwm:

        {
			struct ftop_set_pwm *c = (void*) buffer;
			uint16_t npwm = c->pwm >= 512 ? 512 : c->pwm;

			if ((c->devid >= ftdev_s1) && (c->devid <= ftdev_s3)) {
				HWServo_SetDuty(c->devid - ftdev_s1, npwm);
			} else if (c->devid >= ftdev_o1 && c->devid <= ftdev_o8) {
				uint8_t cpwm = npwm >= 512 ? 255 : npwm / 2;

				OS_LOCK(drv8908_mutex);
				int i = c->devid - ftdev_o1;
				txt2_server.current.motormode &= ~(1 << (i / 2));
				txt2_server.current.opwm[i] = npwm;
				DRV8908_set_pwm(i, cpwm);
				OS_UNLOCK(drv8908_mutex);
			}
        }
		break;

    case ftop_op_zero:
		{
			zops++;
			struct ftop_set_pwm *c = (void*) buffer;
			if (c->devid >= ftdev_c1 && c->devid <= ftdev_c4) { /* C1-C4 */
				int i = c->devid - ftdev_c1;
				txt2_server.fcounters[i] = txt2_server.counters[i];
			} else if (c->devid >= ftdev_m1 && c->devid <= ftdev_m4) { /* M1-M4 */
				int i = (c->devid - ftdev_m1) / 2;
				txt2_server.fcounters[i] = txt2_server.counters[i];
			}
		}
		break;

    case ftop_op_disconnect:

    	{
			for (int i = 0; i < FT_MAX_TXT; i++) {
				ftc->opened[i] = 0; /* TODO */
				ftc->usage[i] = 0;
				if (i && (txt2_server.current.common.mode == ft_txt_master)) {
					UpdateRemoteConfig(ftc, &txt2_server.slaves[i - 1],
							&zero_config, i, 0);
				} else {
					ProcessNewConfig(0, &zero_config);
				}
			}
			txt2_server.stat.clients &= ~ (1 << ftc->id);
		}
		break;

    case ftop_op_terminator:

		{
			struct ftop_set_mode *smode =  (void*)buffer;

			HWFDCAN_set_terminator(smode->mode);
		}
		break;

    case ftop_op_set_bitrate:

		{
			struct ftop_set_mode *smode =  (void*)buffer;

			HWFDCAN_set_bitrate(smode->mode);
		}
		break;

    case ftop_op_set_uid:

		{
			struct ftop_set_mode *smode =  (void*)buffer;

			buid = smode->mode;
		    txt2_server.uids[0] = buid;
		    txt2_server.current.common.uid = txt2_server.uids[0];
			GenerateShortName();
			sprintf(txt2_server.current.common.sn, "ft%8.8lX", txt2_server.uids[0]);
		}
		break;

    case ftop_op_set_name:

    	{
			struct ftop_set_name *set_name = (void*) buffer;

			memcpy(txt2_server.current.common.device_name, set_name->name,
					sizeof(txt2_server.current.common.device_name));
		}
		break;

    case ftop_op_power_down:

		PowerDown();
		break;

    case ftop_op_reset:
		/* TODO: */
		break;

    default: /* Something wrong, close connection */
	    //OS_DEBUG("Error cmd %d\n", ftcmd->cmd)
        break;
    }
}

/**
 * @brief Procedure to calculate resources usage for all clients and master
 * @param ntxt TXT2 controller number, 0 - local
 * @return Resources usage for selected TXT2 controller
 **/
uint32_t CalculateTotalHWUsage(int ntxt)
{
	uint32_t usage = master_usage;

	for (int i = 0; i < FT_MAX_CLIENTS; i++) {
		usage |= clients[i].usage[ntxt];
	}

	return usage;
}
