/*
 * Copyright (c) 2020 RTSoft. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @file ftutil/ftcalib.c
 * @brief Calibration utility
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "ftcalib.h"

#define M4_MEM_SIZE (256 * 1024) /*!< M4 memory size */
#define M4_ADDR 0x10000000		 /*!< M4 firmware address */

static int verbose = 0;

static void Usage(int code)
{
	fprintf(stderr, "Usage: ftconfig command\n"
			        "Commands:\n"
					"    show - show parameters\n"
					"    adc d1 r1 d2 r2 - common channels calibration\n"
					"    adcs | adct1 | adct2 | adcdiv | adcr1 | adcr2 | dt | rt | u0t | u1t | ut | ct | cnr | dnr value - set value\n"
			        "    tm0-tm4 - set tm value\n"
					"    raw  - reset all parameters\n"
					"    areset - reset ADC calibration coefficients\n"
					"    amult float - multiply analog sampling delay\n"
					"    chan chanid d1 r1 d2 r2 - channel chanid (1-8) calibration\n"
					"    save file - save parameters to file\n"
					"    load file - load parameters from file\n"
					"    fix file  - fix M4 firmware file\n"
            );
    exit(code);
}


static void Show(struct ftcalib *ftc)
{
	if (verbose)
		fprintf(stderr, "Mark %p\n", (void*)ftc);

	fprintf(stderr, "Version 0x%x\n", ftc->version);
	if (ftc->version != FTCALIB_VER) {
		fprintf(stderr, "Unsupported verison\n");
		return;
	}
	fprintf(stderr, "UID     0x%x\n", ftc->uid);
	fprintf(stderr, "ADCS    %d\n", ftc->adcs);
	fprintf(stderr, "ADCT1   %d ns\n", ftc->adct1);
	fprintf(stderr, "ADCT2   %d ns\n", ftc->adct2);
	fprintf(stderr, "U0      %d ns\n", ftc->u0t);
	fprintf(stderr, "U1      %d ns\n", ftc->u1t);
	fprintf(stderr, "UT      %d ns\n", ftc->ut);
	fprintf(stderr, "DT      %d ns\n", ftc->dt);
	fprintf(stderr, "RT      %d ms\n", ftc->rt);
	fprintf(stderr, "CNR     %d\n", ftc->cnr);
	fprintf(stderr, "DNR     %d\n", ftc->dnr);
	fprintf(stderr, "CT      %d ms\n", ftc->ct);

	fprintf(stderr, "TM      %d %d %d %d %d\n",
			ftc->tm[0], ftc->tm[1], ftc->tm[2], ftc->tm[3], ftc->tm[4]);

	fprintf(stderr, "     ");
	for (int i = 1; i <= 8; i++) {
		fprintf(stderr, " %4d", i);
	}

	fprintf(stderr, "\nDIV1 ");
	for (int i = 0; i < 8; i++) {
		fprintf(stderr, " %4d", ftc->channel[i].div1);
	}

	fprintf(stderr, "\nVAL1 ");
	for (int i = 0; i < 8; i++) {
		fprintf(stderr, " %4d", ftc->channel[i].val1);
	}

	fprintf(stderr, "\nDIV2 ");
	for (int i = 0; i < 8; i++) {
		fprintf(stderr, " %4d", ftc->channel[i].div2);
	}

	fprintf(stderr, "\nVAL2 ");
	for (int i = 0; i < 8; i++) {
		fprintf(stderr, " %4d", ftc->channel[i].val2);
	}

	fprintf(stderr, "\n");
}

#define CALIB_ERROR exit(3) /*!< Standard error processing in ftcalibcommon.h for ftcalib */
#include "ftcalibcommon.h"

/* Save calibration parameters to file */
static void Save(struct ftcalib *ftc, char *fname)
{
	struct ftcalib out = *ftc;
	FILE *tf = fopen(fname, "w");
	if (tf == NULL) {
		perror("fopen");
		exit(3);
	}

	out.mmark = 0;

	if (fwrite(&out, 1, sizeof(out), tf) != sizeof(out)) {
		perror("fwrite");
		exit(3);
	}
	fclose(tf);
}

static void Fix(struct ftcalib *ftc, char *fname)
{
	struct stat s;

	int fd = open(fname, O_RDWR);
	if (fd < 0) {
		perror("open");
		exit(1);
	}
	if (fstat(fd, &s) < 0) {
		perror("fstat");
		exit(1);
	}

	void *xaddr = mmap(NULL, s.st_size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
	if (xaddr == MAP_FAILED) {
		perror("mmap");
		exit(1);
	}

	if (verbose)
		fprintf(stderr, "faddr %p\n", (void*)xaddr);

	struct ftcalib *out = FindMark(xaddr, s.st_size, 0);
	if (verbose)
		fprintf(stderr, "mark %p\n", (void*)out);

	Copy(ftc, out);
	close(fd);
}

/**
 * @brief Standard main procedure
 * @param [in] argc number of arguments
 * @param [in] argv arguments
 * @return Error code
 */
int main(int argc, char **argv)
{
	struct ftcalib *ftc;
	int32_t value;

	if (argv[1] != NULL && !strcmp(argv[1], "-v")) {
		verbose = 1;
		argc--;
		argv++;
	}

	if (argc <= 1)
		Usage(0);

	/* Map M4 memory */
	uint32_t offset = 0;
	int fd = open("/dev/m4mem", O_RDWR | O_SYNC);
	if (fd < 0) {
		perror("m4mem");
		fd = open("/dev/mem", O_RDONLY | O_CLOEXEC);
		offset = M4_ADDR; /* Physical address */
		if (fd < 0)
			return 1;
	}
	void *xaddr = mmap(NULL, M4_MEM_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED, fd, (off_t) offset);
	if (xaddr == MAP_FAILED) {
		perror("mmap");
		return 1;
	}

	if (verbose)
		fprintf(stderr, "Map %p\n", xaddr);

	/* Find mark in memory */
	ftc = FindMark(xaddr, M4_MEM_SIZE, 1);
	if (verbose)
		fprintf(stderr, "mark %p\n", (void*)ftc);

	if (!strcmp(argv[1], "show")) {
		Show(ftc);
	} else if(!strcmp(argv[1], "adcs")) {
		sscanf(argv[2], "%d", &value);
		ftc->adcs = value;
	} else if(!strcmp(argv[1], "adct1")) {
		sscanf(argv[2], "%d", &value);
		ftc->adct1 = value;
	} else if(!strcmp(argv[1], "adct2")) {
		sscanf(argv[2], "%d", &value);
		ftc->adct2 = value;
	} else if(!strcmp(argv[1], "dt")) {
			sscanf(argv[2], "%d", &value);
			ftc->dt = value;
	} else if(!strcmp(argv[1], "cnr")) {
			sscanf(argv[2], "%i", &value);
			ftc->cnr = value;
	} else if(!strcmp(argv[1], "ct")) {
				sscanf(argv[2], "%i", &value);
				ftc->ct = value;
	} else if(!strcmp(argv[1], "dnr")) {
			sscanf(argv[2], "%i", &value);
			ftc->dnr = value;
	} else if(!strcmp(argv[1], "rt")) {
			sscanf(argv[2], "%d", &value);
			ftc->rt = value;
	} else if(!strcmp(argv[1], "u0t")) {
			sscanf(argv[2], "%d", &value);
			ftc->u0t = value;
	} else if(!strcmp(argv[1], "u1t")) {
			sscanf(argv[2], "%d", &value);
			ftc->u1t = value;
	} else if(!strcmp(argv[1], "ut")) {
			sscanf(argv[2], "%d", &value);
			ftc->ut = value;
	} else if(!strcmp(argv[1], "adcd1")) {
		sscanf(argv[2], "%d", &value);
		for (int i = 0; i < 8; i++) ftc->channel[i].div1 = value;
	} else if(!strcmp(argv[1], "adcd2")) {
		sscanf(argv[2], "%d", &value);
		for (int i = 0; i < 8; i++) ftc->channel[i].div2 = value;
	} else if(!strcmp(argv[1], "adcr1")) {
		sscanf(argv[2], "%d", &value);
		for (int i = 0; i < 8; i++) ftc->channel[i].val1 = value;
	} else if(!strcmp(argv[1], "adcr2")) {
		sscanf(argv[2], "%d", &value);
		for (int i = 0; i < 8; i++) ftc->channel[i].val2 = value;
	} else if(!strcmp(argv[1], "save")) {
		Save(ftc, argv[2]);
	} else if(!strcmp(argv[1], "load")) {
		LoadCalibFile(ftc, argv[2]);
	} else if(!strcmp(argv[1], "raw")) {
		int32_t rtime = 7000000; /* 7 ms */
		if (argv[2] != NULL)
			sscanf(argv[2], "%i", &rtime);
		ftc->adct1 = rtime;
		ftc->adct2 = rtime;
		for (int i = 0; i < 8; i++) {
			ftc->channel[i].div1 = 1;
			ftc->channel[i].div2 = 1;
			ftc->channel[i].val1 = 1;
			ftc->channel[i].val2 = 1;
		}
	} else if(!strcmp(argv[1], "areset")) {
		for (int i = 0; i < 8; i++) {
			ftc->channel[i].div1 = 1;
			ftc->channel[i].div2 = 1;
			ftc->channel[i].val1 = 1;
			ftc->channel[i].val2 = 1;
		}
	} else if(!strcmp(argv[1], "adc")) { /* d1 r1 d2 r2 */
		sscanf(argv[2], "%d", &value);
		for (int i = 0; i < 8; i++) ftc->channel[i].div1 = value;
		sscanf(argv[3], "%d", &value);
		for (int i = 0; i < 8; i++) ftc->channel[i].val1 = value;
		sscanf(argv[4], "%d", &value);
		for (int i = 0; i < 8; i++) ftc->channel[i].div2 = value;
		sscanf(argv[5], "%d", &value);
		for (int i = 0; i < 8; i++) ftc->channel[i].val2 = value;
	} else if (!strcmp(argv[1], "chan")) { /* i d1 r1 d2 r2 */
		int i;
		sscanf(argv[2], "%d", &i);
		if (i < 1 || i > 8) {
			fprintf(stderr, "Incorrect index\n");
			return 1;
		}
		i--;
		sscanf(argv[3], "%d", &value);
		ftc->channel[i].div1 = value;
		sscanf(argv[4], "%d", &value);
		ftc->channel[i].val1 = value;
		sscanf(argv[5], "%d", &value);
		ftc->channel[i].div2 = value;
		sscanf(argv[6], "%d", &value);
		ftc->channel[i].val2 = value;
	} else if (!strcmp(argv[1], "amult")) { /* float multiplier */
		double v = 10.;
		sscanf(argv[2], "%lf", &v);
		value = ftc->adct1 * v + 199;
		value -= value % 200;
		ftc->adct1 = value;
		value = ftc->adct2 * v + 199;
		value -= value % 200;
		ftc->adct2 = value;
	} else if(!strcmp(argv[1], "fix")) { /* fix firmware */
		Fix(ftc, argv[2]);
	} else if(!strncmp(argv[1], "tm", 2) && argv[1][2] >= '0' && argv[1][2] <= '4' && !argv[1][3]) {
		sscanf(argv[2], "%d", &value);
		ftc->tm[argv[1][2] - '0'] = value;
	} else {
		fprintf(stderr, "Unknown command\n");
		Usage(1);
	}
	close(fd);
	return 0;
}

