/*
 * Copyright (c) 2020 RTSoft. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @file HWCounters.c
 * @brief Procedures to process GPIO for C1-C4
 */

/*
 * Use GPIO for counters because 1 timer can use one clock source
 */

#include "main.h"
#include "cmsis_os.h"
#include "ftshmem.h"

/* Private variables */

static GPIO_InitTypeDef  GPIO_InitStruct;

static int init = 0;

static uint16_t gpiob;
#if TXT2_BOARD_REVISION == 1
static uint16_t gpioe;
#else
static uint16_t gpioa;
#endif

/* External variables */

#if TXT2_BOARD_REVISION == 1
extern int firstMotorCheck;     /*!< 0 - wait nonzero, 1 - wait counter, 2 - wait zero */
extern uint32_t motorCheckTick; /*!< initial nonzero time for M1 PWM motor */
static int motorMaxCounter;
#endif

/* Exported functions */

/**
 * @brief C1-C4 GPIO initialization
 */
void HWCountersInit()
{
    /* Init counters pins */
	GPIO_InitStruct.Pin = COUNTER_1_Pin;
	GPIO_InitStruct.Mode = GPIO_MODE_INPUT;
	GPIO_InitStruct.Pull = GPIO_NOPULL;
	GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_VERY_HIGH;
	HAL_GPIO_Init(COUNTER_1_GPIO_Port, &GPIO_InitStruct);

	GPIO_InitStruct.Pin = COUNTER_2_Pin;
	GPIO_InitStruct.Mode = GPIO_MODE_INPUT;
	GPIO_InitStruct.Pull = GPIO_NOPULL;
	GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_VERY_HIGH;
	HAL_GPIO_Init(COUNTER_2_GPIO_Port, &GPIO_InitStruct);

	GPIO_InitStruct.Pin = COUNTER_3_Pin;
	GPIO_InitStruct.Mode = GPIO_MODE_INPUT;
	GPIO_InitStruct.Pull = GPIO_NOPULL;
	GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_VERY_HIGH;
	HAL_GPIO_Init(COUNTER_3_GPIO_Port, &GPIO_InitStruct);

	GPIO_InitStruct.Pin = COUNTER_4_Pin;
	GPIO_InitStruct.Mode = GPIO_MODE_INPUT;
	GPIO_InitStruct.Pull = GPIO_NOPULL;
	GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_VERY_HIGH;
	HAL_GPIO_Init(COUNTER_4_GPIO_Port, &GPIO_InitStruct);

	gpiob = GPIOB->IDR;
#if TXT2_BOARD_REVISION == 1
	gpioe = GPIOE->IDR;
#else
	gpioa = GPIOA->IDR;
#endif
	init = 1;
}

/**
 * @brief C1-C4 GPIO processing for each 1 ms
 */
void HWCountersUpdate(uint32_t *counters, uint8_t *cstate)
{
	if (!init)
		return;

	uint16_t ngpiob = GPIOB->IDR;
	uint16_t tgpiob = ngpiob & ~gpiob;
#if TXT2_BOARD_REVISION == 1
	uint16_t ngpioe = GPIOE->IDR;
	uint16_t tgpioe = ngpioe & ~gpioe;
	if (tgpioe & COUNTER_1_Pin) {
		counters[0]++;

		if (firstMotorCheck == 1) {
			uint32_t diff = HAL_GetTick() - motorCheckTick;
			if (diff > motorMaxCounter) {
				ftstat->m0maxtime = motorMaxCounter = diff;
			}
			firstMotorCheck = 2;
		}
	}
	if (tgpiob & COUNTER_2_Pin)
		counters[1]++;
	if (tgpiob & COUNTER_3_Pin)
		counters[2]++;
	if (tgpiob & COUNTER_4_Pin)
		counters[3]++;
	cstate[0] = (ngpioe & COUNTER_1_Pin) != 0;
	cstate[1] = (ngpiob & COUNTER_2_Pin) != 0;
	cstate[2] = (ngpiob & COUNTER_3_Pin) != 0;
	cstate[3] = (ngpiob & COUNTER_4_Pin) != 0;
	gpioe = ngpioe;
#else
	uint16_t ngpioa = GPIOA->IDR;
	uint16_t tgpioa = ngpioa & ~gpioa;
	if (tgpiob & COUNTER_1_Pin)
		counters[0]++;
	if (tgpiob & COUNTER_2_Pin)
		counters[1]++;
	if (tgpioa & COUNTER_3_Pin)
		counters[2]++;
	if (tgpioa & COUNTER_4_Pin)
		counters[3]++;
	cstate[0] = (ngpiob & COUNTER_1_Pin) != 0;
	cstate[1] = (ngpiob & COUNTER_2_Pin) != 0;
	cstate[2] = (ngpioa & COUNTER_3_Pin) != 0;
	cstate[3] = (ngpioa & COUNTER_4_Pin) != 0;
	gpioa = ngpioa;
#endif
	gpiob = ngpiob;

}
