/*
 * Copyright (c) 2020 RTSoft. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @file HWLoop.c
 * @brief FSM to check all generic inputs I1-I8
 */

#include "HWLoop.h"
#include "main.h"
#include "ftshmem.h"
#include "ftcalib.h"

#include "cmsis_os.h"
#include <semphr.h>
#include "openamp_log.h"

#define USE_DEBUG_OUTPUT 0

extern ADC_HandleTypeDef hadc1;		/*!< ADC1 HAL handle */
extern UART_HandleTypeDef huart6;	/*!< USUART6 HAL handle */
extern struct ftcalib ftcalib;		/*!< Common calibration values */

enum { DI_GPIO_mode, DI_U_mode};

static int dimode = -1;
#define ULTRA_TIMEOUT_MIN 16				/*!< Minimal number of timeoouts for ultrasonic */
#define ULTRA_PERIOD 10						/*!< Time between ultrasonic starts in ms */
static int ultratimeout[FT_MAX_INPUTS];		/*!< Current ultrasonic timeout value per channel */
static uint32_t ultratime[FT_MAX_INPUTS];   /*!< Previous ultrasonic read time to select channel */
static int us_select;                       /*!< Current selected ultrasonic */
static GPIO_InitTypeDef  GPIO_InitStruct;
static int debug_isr = 0;

/* Interrupt context */
enum {
	nostate,     /* Select new input */
	dwait_state, /* Wait digital input */
	await_state, /* Wait analog input */
	uzero_state, /* Wait zero byte output */
	uone_state,  /* Output stop bit */
	uwait_state  /* Wait next byte input */
};
static int volatile sstate; /* Input machine state */

static int i18selector = 0;              /* Current input               */
static struct ft_inputs_config *sinputs; /* Inputs configuration        */
static ftvalues *svalues;                /* Plave for values in shared memory */

static uint16_t rawdata[FT_MAX_INPUTS]; /* raw data from inputs         */
static uint8_t rawchar[3];              /* Ultrasonic characters        */
static int ichar;                       /* Current ultrasonic character */

static StaticSemaphore_t hwEventSemaphore;
static SemaphoreHandle_t hwEventHandle;

/* Prepared time values for timer registers */
static struct {
	uint32_t adct1;
	uint32_t adct2;
	uint32_t u0t;
	uint32_t u1t;
	uint32_t ut;
	uint32_t dt;
} rvalues;

#define PAIR(n) {(n & 1) ? (IOMUX_A_Pin << 16) : IOMUX_A_Pin, \
						   ((n & 2) ? (IOMUX_B_Pin << 16) : IOMUX_B_Pin) \
						|  ((n & 4) ? (IOMUX_C_Pin << 16) : IOMUX_C_Pin)} \
						/*!< macro to initialize group of pins */

static uint32_t pins[8][2] = { PAIR(0), PAIR(1), PAIR(2), PAIR(3),
		PAIR(4), PAIR(5), PAIR(6), PAIR(7) };

/*
 * Multiplexor pins:
 * TXT2.0 TXT2.1
 * PC4    PC4  ADC1_INP4 Analog input
 * PD14   PD14 IOMUX_A
 * PG7    PG7  IOMUX_DO
 * PG9    PG9  IOMUX_B
 * PG10   PG10 IOMUX_C
 * PG12   PG12 IOMUX_EN
 * PG13   PG13 AIO_PRESCALE (use 1 by default)
 * PG14   PG14 IOMUX_DI
 *
 * Propogation time for CD4051 <= 720 ns
 */

/**
 * @brief I1-I8 initialization procedure
 */
void HWLoop_Init()
{
	/* Logical structures init */
	hwEventHandle = xSemaphoreCreateBinaryStatic(&hwEventSemaphore);

	/* Init all GPIO pins */

	/* Init mux pins */
	GPIO_InitStruct.Pin = IOMUX_A_Pin;
	GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_PP;
	GPIO_InitStruct.Pull = GPIO_NOPULL;
	GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_VERY_HIGH;
	HAL_GPIO_Init(IOMUX_A_GPIO_Port, &GPIO_InitStruct);

	GPIO_InitStruct.Pin = IOMUX_B_Pin;
	GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_PP;
	GPIO_InitStruct.Pull = GPIO_NOPULL;
	GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_VERY_HIGH;
	HAL_GPIO_Init(IOMUX_B_GPIO_Port, &GPIO_InitStruct);

	GPIO_InitStruct.Pin = IOMUX_C_Pin;
	GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_PP;
	GPIO_InitStruct.Pull = GPIO_NOPULL;
	GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_VERY_HIGH;
	HAL_GPIO_Init(IOMUX_C_GPIO_Port, &GPIO_InitStruct);

	GPIO_InitStruct.Pin = IOMUX_DO_Pin;
	GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_PP;
	GPIO_InitStruct.Pull = GPIO_NOPULL;
	GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_VERY_HIGH;
	HAL_GPIO_Init(IOMUX_DO_GPIO_Port, &GPIO_InitStruct);

	GPIO_InitStruct.Pin = IOMUX_EN_Pin;
	GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_PP;
	GPIO_InitStruct.Pull = GPIO_NOPULL;
	GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_VERY_HIGH;
	HAL_GPIO_Init(IOMUX_EN_GPIO_Port, &GPIO_InitStruct);
	HAL_GPIO_WritePin(IOMUX_EN_GPIO_Port, IOMUX_EN_Pin, GPIO_PIN_SET);

	GPIO_InitStruct.Pin = AIO_PRESCALE_Pin;
	GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_PP;
	GPIO_InitStruct.Pull = GPIO_NOPULL;
	GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_VERY_HIGH;
	HAL_GPIO_Init(AIO_PRESCALE_GPIO_Port, &GPIO_InitStruct);

	GPIO_InitStruct.Pin = IOMUX_DI_Pin;
	GPIO_InitStruct.Mode = GPIO_MODE_INPUT;
	GPIO_InitStruct.Pull = GPIO_NOPULL;
	GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_VERY_HIGH;
	HAL_GPIO_Init(IOMUX_DI_GPIO_Port, &GPIO_InitStruct);
	dimode = DI_GPIO_mode;
}

#if USE_DEBUG_OUTPUT && (TXT2_BOARD_REVISION == 1)
static uint32_t t7start;
static uint32_t t7ticks;
#endif

static void TIM7Start(int out, uint32_t ticks)
{
	/* Reset state */
	TIM7->DIER &= ~(TIM_IT_UPDATE);
	TIM7->SR = ~TIM_IT_UPDATE;
	TIM7->CR1 &= ~(TIM_CR1_CEN);

#if USE_DEBUG_OUTPUT && (TXT2_BOARD_REVISION == 1)
	if (out) {
		uint32_t t7diff = TIM5->CNT - t7start;
		log_info("t7out %ld -> %ld : %ld\n", t7ticks, t7diff, TIM7->CNT);
	}
#endif

	/* Start new */
	TIM7->ARR = ticks;
	TIM7->DIER |= TIM_IT_UPDATE;
#if USE_DEBUG_OUTPUT && (TXT2_BOARD_REVISION == 1)
	t7ticks = ticks;
	t7start = TIM5->CNT;
	if (!out) {
		log_info("t7start %ld -> %ld : %ld\n", t7start, t7ticks, TIM7->CNT);
	}
#endif
	TIM7->CNT = 0;
	TIM7->CR1 |= TIM_CR1_CEN;
}

static void TIM7Stop()
{
	TIM7->DIER &= ~(TIM_IT_UPDATE);
	TIM7->CR1 &= ~(TIM_CR1_CEN); /* HAL_TIM_Base_Stop_IT(&htim7); */
	TIM7->SR = ~TIM_IT_UPDATE;
#if USE_DEBUG_OUTPUT && (TXT2_BOARD_REVISION == 1)
	uint32_t t7diff = TIM5->CNT - t7start;
	log_info("t7stop %ld -> %ld\n", t7ticks, t7diff);
#endif
}

/**
 * @brief Prepare simple inpit I1-I8 initialization in cycle
 * @param mode mode for input
 */
static void HWStartSimple(int mode)
{

	if (dimode != DI_GPIO_mode) {
		uint32_t temp = IOMUX_DI_GPIO_Port->MODER;
		int position = NPIN(IOMUX_DI_Pin);
		temp &= ~(GPIO_MODER_MODER0 << (position * 2));
		temp |= ((GPIO_MODE_INPUT & 3) << (position * 2)); /* GPIO_MODE == 3 */
		IOMUX_DI_GPIO_Port->MODER = temp;

#if 0 /* Value is stable */
		temp = IOMUX_DI_GPIO_Port->PUPDR;
		temp &= ~(GPIO_PUPDR_PUPDR0 << (position * 2));
		temp |= GPIO_NOPULL << (position * 2);
		IOMUX_DI_GPIO_Port->PUPDR = temp;
#endif
		dimode = DI_GPIO_mode;
	}

	uint32_t bsrr = IOMUX_DO_Pin << 16; /* GPIO_NUMBER == 16 */
	/* Inhibit for active */
	bsrr |= (mode & ftinp_passive) ? IOMUX_EN_Pin : (IOMUX_EN_Pin << 16);
    bsrr |= (mode & ftinp_r2) ?  AIO_PRESCALE_Pin : ( AIO_PRESCALE_Pin << 16);
    IOMUX_DO_GPIO_Port->BSRR = bsrr;
}

/* Ultrasonic bytes :
 * 0: 1 Valid# W/D# ID1 ID0 L9 L8 L7
 * 1: 0 L6 L5 L4 L3 L2 L1 L0
 */

/**
 * @brief Prepare I1-I8 input for ultrasonic in cycle
 * @param mode mode for input
 */
static void HWStartUltra()
{
	IOMUX_DO_GPIO_Port->BSRR = IOMUX_EN_Pin | (IOMUX_DO_Pin << 16);
	if (dimode != DI_U_mode) {
		int position = NPIN(IOMUX_DI_Pin);

		/* Alternate function mode selection */
		uint32_t temp = IOMUX_DI_GPIO_Port->AFR[position >> 3];
		temp &= ~((uint32_t)0xF << ((uint32_t)(position & (uint32_t)0x07) * 4)) ;
		temp |= ((uint32_t)(GPIO_AF7_USART6) << (((uint32_t)position & (uint32_t)0x07) * 4));
		IOMUX_DI_GPIO_Port->AFR[position >> 3] = temp;

	    /* Configure IO Direction mode (Input, Output, Alternate or Analog) */
	    temp = IOMUX_DI_GPIO_Port->MODER;
	    temp &= ~(GPIO_MODER_MODER0 << (position * 2));
	    temp |= ((GPIO_MODE_AF_OD & 3) << (position * 2));
	    IOMUX_DI_GPIO_Port->MODER = temp;

#if 0 /* Values are stable */
        temp = IOMUX_DI_GPIO_Port->OSPEEDR;
        temp &= ~(GPIO_OSPEEDR_OSPEEDR0 << (position * 2));
        temp |= (GPIO_SPEED_FREQ_LOW << (position * 2));
        IOMUX_DI_GPIO_Port->OSPEEDR = temp;

        /* Configure the IO Output Type */
        temp = IOMUX_DI_GPIO_Port->OTYPER;
        temp &= ~(GPIO_OTYPER_OT0 << position);
        temp |= (((GPIO_MODE_AF_OD & 0x10) >> 4) << position);
        IOMUX_DI_GPIO_Port->OTYPER = temp;

        /* Activate the Pull-up or Pull down resistor for the current IO */
        temp = IOMUX_DI_GPIO_Port->PUPDR;
        temp &= ~(GPIO_PUPDR_PUPDR0 << (position * 2));
        temp |= (GPIO_NOPULL << (position * 2));
        IOMUX_DI_GPIO_Port->PUPDR = temp;
#endif

        AIO_PRESCALE_GPIO_Port->BSRR = AIO_PRESCALE_Pin << 16;

		dimode = DI_U_mode;
	}
	ichar = 0;
}

/**
 * @brief Start next I1-I8 input
 * @param emode mode for input: 0 - usual call, 1 - call from interrupt
 */
int HWLoopStartNext(int emode)
{
	sstate = nostate;
	if (debug_isr) {
		//log_info("debug_isr 0x%x\n", debug_isr);
		debug_isr = 0;
		//log_info("rawchar %d 0x%x 0x%x 0x%x\n", ichar, rawchar[0], rawchar[1], rawchar[2]);
	}

    for(; i18selector < FT_MAX_INPUTS; i18selector++) {
		int mode = sinputs->inputs[i18selector][0];

		if (mode == ftinp_unknown) {
			svalues->inputs[i18selector] = 0;
			continue;
		}

        /* Setup mux */
		uint32_t *p = &pins[i18selector][0];
		IOMUX_A_GPIO_Port->BSRR = *p++;
		IOMUX_C_GPIO_Port->BSRR = *p;

		if (mode & ftinp_ultrasonic) { /* start and wait 3 bytes from UART */
			if (i18selector != us_select)
				continue; /* Skip in this cycle */

			ultratime[i18selector] = HAL_GetTick();

			HWStartUltra();
			sstate = uone_state;
			TIM7Start(0, rvalues.u1t);
		} else if (mode & ftinp_analog) { /* start and wait ADC */
			sstate = await_state;
			HWStartSimple(mode);
			HAL_ADC_Start_IT(&hadc1);
			if (mode & ftinp_r2) {
				TIM6->ARR = rvalues.adct2;
			} else {
				TIM6->ARR = rvalues.adct1;
			}
			TIM6->CNT = 0;
			TIM6->CR1 |= TIM_CR1_CEN; /* HAL_TIM_Base_Start(&htim6); */
		} else { /* start and wait digital */
			sstate = dwait_state;
			HWStartSimple(mode);
			TIM7Start(0, rvalues.dt);
		}
		return 1;
    }

    if (emode == 1) {  /* send event to thread */
        xSemaphoreGiveFromISR(hwEventHandle, NULL);
    }

    return 0; // No reason for wait
}

/**
 * @brief Process digital interrupt from I1-I8 input, read value and start next
 */
void TIM7_IRQHandler(void)
{
	ftstat->tim7counter++;
	TIM7->SR = ~TIM_IT_UPDATE; /* __HAL_TIM_CLEAR_IT(&htim7, TIM_IT_UPDATE); */

	if(hwEventHandle == NULL) {
		return;
	}

	if (sstate == uzero_state) {
	    /* Enable the UART Data Register Not Empty interrupt */
		debug_isr |= READ_REG(USART6->ISR);
		/* Clear errors */
		if (READ_REG(USART6->ISR) & USART_ISR_FE) {
			USART6->ICR = UART_CLEAR_FEF; /* __HAL_UART_CLEAR_FEFLAG(&huart6); */
		}
		/* Clear buffer */
		if (READ_REG(USART6->ISR) & USART_ISR_RXNE_RXFNE) {
			USART6->RQR = UART_RXDATA_FLUSH_REQUEST;
		}

		IOMUX_EN_GPIO_Port->BSRR = IOMUX_DO_Pin << 16;
		SET_BIT(USART6->CR1, USART_CR1_RXNEIE_RXFNEIE | USART_CR1_PEIE | USART_CR1_RE | USART_CR1_UE);
		sstate = uwait_state;
		TIM7Start(1, rvalues.ut);
		return;
	}
	if (sstate == uone_state) {
		IOMUX_DO_GPIO_Port->BSRR = IOMUX_DO_Pin | IOMUX_EN_Pin;

		sstate = uzero_state;
		TIM7Start(1, rvalues.u0t);
		return;
	}
	if (sstate == uwait_state) {
		/* Timeout */
		CLEAR_BIT(huart6.Instance->CR1,
				(USART_CR1_PEIE | USART_CR1_RXNEIE_RXFNEIE | USART_CR1_RE | USART_CR1_UE));
		//log_info("timeout isr 0x%lx\n", USART6->ISR);
		USART6->ICR = USART_ICR_ORECF | USART_ICR_NECF | USART_ICR_FECF
				| USART_ICR_PECF;
		ultratimeout[i18selector]++;
		if (ultratimeout[i18selector] > ULTRA_TIMEOUT_MIN) {
			rawdata[i18selector] = 0xffff;
		}
		TIM7Stop();
		i18selector++;
		HWLoopStartNext(1);
		return;
	}

	if (sstate == dwait_state) {
		rawdata[i18selector] = (IOMUX_DI_GPIO_Port->IDR >> NPIN(IOMUX_DI_Pin)) & 1;
		TIM7Stop();

		i18selector++;
		HWLoopStartNext(1);
		return;
	}
}

/**
  * @brief This function handles ADC1 global interrupt.
  * Process analog interrupt from I1-I8 input, read value and start next
  */
void ADC1_IRQHandler(void)
{
	ftstat->adccounter++;

	/* __HAL_ADC_CLEAR_FLAG(&hadc1, (ADC_FLAG_EOC | ADC_FLAG_EOS)) */
	ADC1->ISR = ADC_FLAG_EOC | ADC_FLAG_EOS;

	if(hwEventHandle == NULL) {
		return;
	}

	if (sstate == await_state) {
		rawdata[i18selector] = ADC1->DR; /* HAL_ADC_GetValue(&hadc1); */

		i18selector++;
		HWLoopStartNext(1);
	}
}

/**
 * @brief Process ultrasonic interrupt from I1-I8 input, read value and start next
 */
void USART6_IRQHandler(void)
{

	ftstat->usart6counter++;

	debug_isr |= READ_REG(USART6->ISR);

	if(hwEventHandle == NULL) {
		return;
	}

	if (READ_REG(USART6->ISR) & USART_ISR_FE) {
		USART6->ICR = UART_CLEAR_FEF; /* __HAL_UART_CLEAR_FEFLAG(&huart6); */
	}

 	while ((ichar < 3) && (READ_REG(USART6->ISR) & USART_ISR_RXNE_RXFNE)) {
 		debug_isr |= READ_REG(USART6->ISR);
		/* Read next byte */
		uint16_t rdr = READ_REG(USART6->RDR);

		if (rdr == 0xff) { // drop it
			ichar = 0;
			continue;
		}

		if (rdr & 0x80) { /* second byte */
			ichar = 0;
			rawchar[ichar++] = 0;
		}
		rawchar[ichar++] = rdr;
	}

	if (ichar >= 3) {
		/* Stop Tim7 */
		TIM7Stop();

		CLEAR_BIT(USART6->CR1,
				(USART_CR1_PEIE | USART_CR1_RXNEIE_RXFNEIE | USART_CR1_RE | USART_CR1_UE));
		USART6->ICR = USART_ICR_ORECF | USART_ICR_NECF | USART_ICR_FECF
				| USART_ICR_PECF;

		/* Convert to value */
		if (rawchar[0] == 0 && ((rawchar[1] & 0xf8) == 0x80) && !(rawchar[2] & 0x80)) {
			rawdata[i18selector] = ((rawchar[1] & 0x7) << 7) | rawchar[2];
			ultratimeout[i18selector] = 0;
		}
		i18selector++;
		HWLoopStartNext(1);
	}
}

/**
 * @brief Main procedure for I1-I8 inputs, start new cycle and process values
 * @param values buffer for new values
 * @param inputs configuration for I1-I8
 */
void HWLoop(ftvalues *values, struct ft_inputs_config *inputs)
{
	i18selector = 0;
	svalues = values;
	sinputs = inputs;

    /* Can be optimized */
	rvalues.adct1 = ftcalib.adct1 / I1I8TimeResolution;
	rvalues.adct2 = ftcalib.adct2 / I1I8TimeResolution;
	rvalues.dt    = ftcalib.dt / I1I8TimeResolution;

	/* Select next ultrasonic sensor */
	us_select = -1;
	uint32_t maxt = ULTRA_PERIOD;
	uint32_t tick = HAL_GetTick();

	for (int i = 0; i < FT_MAX_INPUTS; i++) {
		int mode = inputs->inputs[i][0];
		if (!(mode & ftinp_ultrasonic))
			continue;

		uint32_t ct = tick - ultratime[i];
		if (ct <= maxt)
			continue;

		us_select = i;
		maxt = ct;
	}

	if (us_select != -1) {
		rvalues.u0t   = ftcalib.u0t / I1I8TimeResolution;
		rvalues.u1t   = ftcalib.u1t / I1I8TimeResolution;
		rvalues.ut    = ftcalib.ut / I1I8TimeResolution;
	}

	/* Read all raw values */
	if (HWLoopStartNext(0)) {
		/* Wait result */
		xSemaphoreTake(hwEventHandle, 100);
	}

	/* Main loop to convert raw values for all I1-I8 devices */
	for (int i = 0; i < FT_MAX_INPUTS; i++) {
		int mode = inputs->inputs[i][0];

		if (mode == ftinp_unknown) {
			values->inputs[i] = 0;
		} else if (mode & ftinp_ultrasonic) {
			values->inputs[i] = rawdata[i];
		} else if (mode & ftinp_analog) {
			int raw = rawdata[i];
			struct ftcalib_group *fg = &ftcalib.channel[i];

			if (mode & ftinp_r2) {
				values->inputs[i] = raw * fg->val2 / fg->div2;
			} else {
				values->inputs[i] = raw * fg->val1 / fg->div1;
			}
		} else {
			values->inputs[i] = rawdata[i];
		}
	}

}

